# originally from sparkmagic package
# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.s

from dbnd._core.errors import DatabandConfigError
from dbnd._core.utils.http.constants import CONFIGURABLE_RETRY, LINEAR_RETRY


# TODO implement real configuration
# implement configuration per name
conf_retry_seconds_to_sleep_list = [0.2, 0.5, 1, 3, 5]
conf_retry_policy_max_retries = 5


def get_retry_policy(name, policy=None, seconds_to_sleep=5, max_retries=5):
    # provide policy per name
    policy = policy or LINEAR_RETRY

    if policy == LINEAR_RETRY:
        return LinearRetryPolicy(
            seconds_to_sleep=seconds_to_sleep, max_retries=max_retries
        )
    elif policy == CONFIGURABLE_RETRY:
        return ConfigurableRetryPolicy(
            retry_seconds_to_sleep_list=conf_retry_seconds_to_sleep_list,
            max_retries=conf_retry_policy_max_retries,
        )
    else:
        raise DatabandConfigError(u"Retry policy '{}' not supported".format(policy))


class LinearRetryPolicy(object):
    """Retry policy that always returns the same number of seconds to sleep between calls,
    takes all status codes 500 or above to be retriable, and retries a given maximum number of times."""

    def __init__(self, seconds_to_sleep, max_retries):
        self._seconds_to_sleep = seconds_to_sleep
        self.max_retries = max_retries

    def should_retry(self, status_code, error, retry_count):
        if None in (status_code, retry_count):
            return False
        if status_code < 500 or error:
            return False
        if self.max_retries != -1 and retry_count >= self.max_retries:
            return False
        return True

    def seconds_to_sleep(self, retry_count):
        return self._seconds_to_sleep


class ConfigurableRetryPolicy(LinearRetryPolicy):
    """Retry policy that returns a configurable number of seconds to sleep between calls,
    takes all status codes 500 or above to be retriable, and retries a given maximum number of times.
    If the retry count exceeds the number of items in the list, last item in the list is always returned."""

    def __init__(self, retry_seconds_to_sleep_list, max_retries):
        super(ConfigurableRetryPolicy, self).__init__(-1, max_retries)

        # If user configured to an empty list, let's make this behave as
        # a Linear Retry Policy by assigning a list of 1 element.
        if len(retry_seconds_to_sleep_list) == 0:
            retry_seconds_to_sleep_list = [5]
        elif not all(n > 0 for n in retry_seconds_to_sleep_list):
            raise DatabandConfigError(
                u"All items in the list in your config need to be positive for configurable retry policy"
            )

        self.retry_seconds_to_sleep_list = retry_seconds_to_sleep_list
        self._max_index = len(self.retry_seconds_to_sleep_list) - 1

    def seconds_to_sleep(self, retry_count):
        index = max(retry_count - 1, 0)
        if index > self._max_index:
            index = self._max_index

        return self.retry_seconds_to_sleep_list[index]



import abc
import logging
import typing

import six

from dbnd._core.configuration.config_store import ConfigMergeSettings
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.constants import _ConfigParamContainer, _TaskParamContainer
from dbnd._core.current import get_databand_context
from dbnd._core.plugin.dbnd_airflow_operator_plugin import (
    build_task_at_airflow_dag_context,
    is_in_airflow_dag_build_context,
)
from dbnd._core.task_build.task_context import (
    TaskContextPhase,
    task_context,
    try_get_current_task,
)
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_meta_factory import (
    TaskMetaFactory,
    TrackedTaskMetaFactory,
)
from dbnd._core.task_build.task_registry import get_task_registry


if typing.TYPE_CHECKING:
    from dbnd._core.task.base_task import _BaseTask

logger = logging.getLogger(__name__)


class TaskMetaclass(abc.ABCMeta):
    """
    The Metaclass of :py:class:`Task`.
    """

    def __new__(mcs, classname, bases, classdict):
        """
        Custom class creation for namespacing.

        Also register all subclasses.

        When the set or inherited namespace evaluates to ``None``, set the task namespace to
        whatever the currently declared namespace is.
        """
        cls = super(TaskMetaclass, mcs).__new__(
            mcs, classname, bases, classdict
        )  # type: typing.Type[_BaseTask]

        # we are starting from "not clean" classdict -> it's deserialization
        if classdict.get("task_definition") is not None:
            return cls

        td = cls.task_definition = TaskDefinition(cls, classdict)

        # now we will assign all params
        set_params = td.class_params if cls.is_tracking_mode else td.all_task_params
        for k, v in six.iteritems(set_params):
            setattr(cls, k, v)

        # every time we see new implementation, we want it to have an priority over old implementation
        # we need to switch to dict() and store history else where
        r = get_task_registry()
        r.register_task(cls)

        return cls

    def _build_task_obj(cls, task_meta):
        return super(TaskMetaclass, cls).__call__(task_meta=task_meta)

    def __call__(cls, *args, **kwargs):
        """
        Custom class instantiation utilizing instance cache.
        """
        _dbnd_disable_airflow_inplace = kwargs.pop(
            "_dbnd_disable_airflow_inplace", False
        )
        if (
            is_in_airflow_dag_build_context()
            and not _ConfigParamContainer.is_type_config(cls)
            and not _dbnd_disable_airflow_inplace
            and not getattr(cls, "_dbnd_decorated_task", False)
        ):
            kwargs = kwargs.copy()
            kwargs["_dbnd_disable_airflow_inplace"] = True
            return build_task_at_airflow_dag_context(
                task_cls=cls, call_args=args, call_kwargs=kwargs
            )

        return cls._create_task(args, kwargs)

    def _create_task(cls, args, kwargs):
        task_definition = cls.task_definition
        # we need to have context initialized before we start to run all logic in config() scope
        with config(
            config_values=task_definition.task_defaults_config_store,
            source=task_definition.task_passport.format_source_name("defaults"),
            merge_settings=ConfigMergeSettings.on_non_exists_only,
        ) as task_config:
            # update config with current class defaults
            # we apply them to config only if there are no values (this is defaults)
            return create_dbnd_task(
                config=task_config,
                new_task_factory=cls._build_task_obj,
                task_cls=cls,
                task_args=args,
                task_kwargs=kwargs,
            )

    @classmethod
    def disable_instance_cache(cls):
        """
        Disables the instance cache.
        """
        cls.__instance_cache = None

    @classmethod
    def clear_instance_cache(cls):
        """
        Clear/Reset the instance cache.
        """
        cls.__instance_cache = {}


def create_dbnd_task(config, new_task_factory, task_cls, task_args, task_kwargs):
    # type:(DbndConfig, Any, Type[_BaseTask], Any, Any, bool)->None
    tracking_mode = task_cls.is_tracking_mode

    task_meta_factory = TrackedTaskMetaFactory if tracking_mode else TaskMetaFactory
    factory = task_meta_factory(
        config=config, task_cls=task_cls, task_args=task_args, task_kwargs=task_kwargs,
    )

    task_meta = factory.create_dbnd_task_meta()

    # If a Task has already been instantiated with the same parameters,
    # the previous instance is returned to reduce number of object instances.
    tic = get_databand_context().task_instance_cache
    task = tic.get_task_obj_by_id(task_meta.obj_key.id)
    if not task or tracking_mode or hasattr(task, "_dbnd_no_cache"):
        task = new_task_factory(task_meta)
        tic.register_task_obj_instance(task)

        # now the task is created - all nested constructors will see it as parent
        with task_context(task, TaskContextPhase.BUILD):
            task._initialize()
            task._validate()
            task.task_meta.config_layer = config.config_layer

        tic.register_task_instance(task)

    parent_task = try_get_current_task()
    if (
        parent_task
        and hasattr(task, "task_id")
        and isinstance(task, _TaskParamContainer)
    ):
        parent_task.task_meta.add_child(task.task_id)

    return task



"""Implementation of HjsonDecoder
"""
from __future__ import absolute_import
import re
import sys
import struct
from .compat import fromhex, b, u, text_type, binary_type, PY3, unichr
from .scanner import HjsonDecodeError

# NOTE (3.1.0): HjsonDecodeError may still be imported from this module for
# compatibility, but it was never in the __all__
__all__ = ['HjsonDecoder']

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL

def _floatconstants():
    _BYTES = fromhex('7FF80000000000007FF0000000000000')
    # The struct module in Python 2.4 would get frexp() out of range here
    # when an endian is specified in the format string. Fixed in Python 2.5+
    if sys.byteorder != 'big':
        _BYTES = _BYTES[:8][::-1] + _BYTES[8:][::-1]
    nan, inf = struct.unpack('dd', _BYTES)
    return nan, inf, -inf

NaN, PosInf, NegInf = _floatconstants()

WHITESPACE = ' \t\n\r'
PUNCTUATOR = '{}[],:'

NUMBER_RE = re.compile(r'[\t ]*(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?[\t ]*')
STRINGCHUNK = re.compile(r'(.*?)([\'"\\\x00-\x1f])', FLAGS)
BACKSLASH = {
    '"': u('"'), '\'': u('\''), '\\': u('\u005c'), '/': u('/'),
    'b': u('\b'), 'f': u('\f'), 'n': u('\n'), 'r': u('\r'), 't': u('\t'),
}

DEFAULT_ENCODING = "utf-8"

def getNext(s, end):
    while 1:
        # Use a slice to prevent IndexError from being raised
        ch = s[end:end + 1]
        # Skip whitespace.
        while ch in WHITESPACE:
            if ch == '': return ch, end
            end += 1
            ch = s[end:end + 1]

        # Hjson allows comments
        ch2 = s[end + 1:end + 2]
        if ch == '#' or ch == '/' and ch2 == '/':
            end = getEol(s, end)
        elif ch == '/' and ch2 == '*':
            end += 2
            ch = s[end]
            while ch != '' and not (ch == '*' and s[end + 1] == '/'):
                end += 1
                ch = s[end]
            if ch != '':
                end += 2
        else:
            break

    return ch, end

def getEol(s, end):
    # skip until eol

    while 1:
        ch = s[end:end + 1]
        if ch == '\r' or ch == '\n' or ch == '':
            return end
        end += 1

def skipIndent(s, end, n):
    ch = s[end:end + 1]
    while ch != '' and ch in " \t\r" and (n > 0 or n < 0):
        end += 1
        n -= 1
        ch = s[end:end + 1]
    return end


def scanstring(s, end, encoding=None, strict=True,
        _b=BACKSLASH, _m=STRINGCHUNK.match, _join=u('').join,
        _PY3=PY3, _maxunicode=sys.maxunicode):
    """Scan the string s for a JSON string. End is the index of the
    character in s after the quote that started the JSON string.
    Unescapes all valid JSON string escape sequences and raises ValueError
    on attempt to decode an invalid string. If strict is False then literal
    control characters are allowed in the string.

    Returns a tuple of the decoded string and the index of the character in s
    after the end quote."""
    if encoding is None:
        encoding = DEFAULT_ENCODING
    chunks = []
    _append = chunks.append
    begin = end - 1
    # callers make sure that string starts with " or '
    exitCh = s[begin]
    while 1:
        chunk = _m(s, end)
        if chunk is None:
            raise HjsonDecodeError(
                "Unterminated string starting at", s, begin)
        end = chunk.end()
        content, terminator = chunk.groups()
        # Content is contains zero or more unescaped string characters
        if content:
            if not _PY3 and not isinstance(content, text_type):
                content = text_type(content, encoding)
            _append(content)
        # Terminator is the end of string, a literal control character,
        # or a backslash denoting that an escape sequence follows
        if terminator == exitCh:
            break
        elif terminator == '"' or terminator == '\'':
            _append(terminator)
            continue
        elif terminator != '\\':
            if strict:
                msg = "Invalid control character %r at"
                raise HjsonDecodeError(msg, s, end)
            else:
                _append(terminator)
                continue
        try:
            esc = s[end]
        except IndexError:
            raise HjsonDecodeError(
                "Unterminated string starting at", s, begin)
        # If not a unicode escape sequence, must be in the lookup table
        if esc != 'u':
            try:
                char = _b[esc]
            except KeyError:
                msg = "Invalid \\X escape sequence %r"
                raise HjsonDecodeError(msg, s, end)
            end += 1
        else:
            # Unicode escape sequence
            msg = "Invalid \\uXXXX escape sequence"
            esc = s[end + 1:end + 5]
            escX = esc[1:2]
            if len(esc) != 4 or escX == 'x' or escX == 'X':
                raise HjsonDecodeError(msg, s, end - 1)
            try:
                uni = int(esc, 16)
            except ValueError:
                raise HjsonDecodeError(msg, s, end - 1)
            end += 5
            # Check for surrogate pair on UCS-4 systems
            # Note that this will join high/low surrogate pairs
            # but will also pass unpaired surrogates through
            if (_maxunicode > 65535 and
                uni & 0xfc00 == 0xd800 and
                s[end:end + 2] == '\\u'):
                esc2 = s[end + 2:end + 6]
                escX = esc2[1:2]
                if len(esc2) == 4 and not (escX == 'x' or escX == 'X'):
                    try:
                        uni2 = int(esc2, 16)
                    except ValueError:
                        raise HjsonDecodeError(msg, s, end)
                    if uni2 & 0xfc00 == 0xdc00:
                        uni = 0x10000 + (((uni - 0xd800) << 10) |
                                         (uni2 - 0xdc00))
                        end += 6
            char = unichr(uni)
        # Append the unescaped character
        _append(char)
    return _join(chunks), end

def mlscanstring(s, end):
    """Scan a multiline string"""

    string = ""
    triple = 0

    # we are at ''' - get indent
    indent = 0
    while 1:
        ch = s[end-indent-1]
        if ch == '\n': break
        indent += 1

    # skip white/to (newline)
    end = skipIndent(s, end + 3, -1)

    ch = s[end]
    if ch == '\n': end = skipIndent(s, end + 1, indent)

    # When parsing multiline string values, we must look for ' characters
    while 1:
        ch = s[end:end + 1]
        if ch == '':
            raise HjsonDecodeError("Bad multiline string", s, end);
        elif ch == '\'':
            triple += 1
            end += 1
            if triple == 3:
                if string[-1] == '\n':
                    string = string[:-1] # remove last EOL
                return string, end
            else:
                continue
        else:
            while triple > 0:
                string += '\''
                triple -= 1

        if ch == '\n':
            string += ch
            end = skipIndent(s, end + 1, indent)
        else:
            if ch != '\r':
                string += ch
            end += 1

def scantfnns(context, s, end):
    """Scan s until eol. return string, True, False or None"""

    chf, begin = getNext(s, end)
    end = begin

    if chf in PUNCTUATOR:
        raise HjsonDecodeError("Found a punctuator character when expecting a quoteless string (check your syntax)", s, end);

    while 1:
        ch = s[end:end + 1]

        isEol = ch == '\r' or ch == '\n' or ch == ''
        if isEol or ch == ',' or \
            ch == '}' or ch == ']' or \
            ch == '#' or \
            ch == '/' and (s[end + 1:end + 2] == '/' or s[end + 1:end + 2] == '*'):

            m = None
            mend = end
            if next: mend -= 1

            if chf == 'n' and s[begin:end].strip() == 'null':
                return None, end
            elif chf == 't' and s[begin:end].strip() == 'true':
                return True, end
            elif chf == 'f' and s[begin:end].strip() == 'false':
                return False, end
            elif chf == '-' or chf >= '0' and chf <= '9':
                m = NUMBER_RE.match(s, begin)

            if m is not None and m.end() == end:
                integer, frac, exp = m.groups()
                if frac or exp:
                    res = context.parse_float(integer + (frac or '') + (exp or ''))
                    if int(res) == res and abs(res)<1e10: res = int(res)
                else:
                    res = context.parse_int(integer)
                return res, end

            if isEol:
                return s[begin:end].strip(), end

        end += 1

def scanKeyName(s, end, encoding=None, strict=True):
    """Scan the string s for a JSON/Hjson key. see scanstring"""

    ch, end = getNext(s, end)

    if ch == '"' or ch == '\'':
        return scanstring(s, end + 1, encoding, strict)

    begin = end
    space = -1
    while 1:
        ch = s[end:end + 1]

        if ch == '':
            raise HjsonDecodeError("Bad key name (eof)", s, end);
        elif ch == ':':
            if begin == end:
                raise HjsonDecodeError("Found ':' but no key name (for an empty key name use quotes)", s, begin)
            elif space >= 0:
                if space != end - 1: raise HjsonDecodeError("Found whitespace in your key name (use quotes to include)", s, space)
                return s[begin:end].rstrip(), end
            else:
                return s[begin:end], end
        elif ch in WHITESPACE:
            if space < 0 or space == end - 1: space = end
        elif ch == '{' or ch == '}' or ch == '[' or ch == ']' or ch == ',':
            raise HjsonDecodeError("Found '" + ch + "' where a key name was expected (check your syntax or use quotes if the key name includes {}[],: or whitespace)", s, begin)
        end += 1

def make_scanner(context):
    parse_object = context.parse_object
    parse_array = context.parse_array
    parse_string = context.parse_string
    parse_mlstring = context.parse_mlstring
    parse_tfnns = context.parse_tfnns
    encoding = context.encoding
    strict = context.strict
    object_hook = context.object_hook
    object_pairs_hook = context.object_pairs_hook
    memo = context.memo

    def _scan_once(string, idx):
        try:
            ch = string[idx]
        except IndexError:
            raise HjsonDecodeError('Expecting value', string, idx)

        if ch == '"' or ch == '\'':
            if string[idx:idx + 3] == '\'\'\'':
                return parse_mlstring(string, idx)
            else:
                return parse_string(string, idx + 1, encoding, strict)
        elif ch == '{':
            return parse_object((string, idx + 1), encoding, strict,
                _scan_once, object_hook, object_pairs_hook, memo)
        elif ch == '[':
            return parse_array((string, idx + 1), _scan_once)

        return parse_tfnns(context, string, idx)

    def scan_once(string, idx):
        if idx < 0: raise HjsonDecodeError('Expecting value', string, idx)
        try:
            return _scan_once(string, idx)
        finally:
            memo.clear()

    def scan_object_once(string, idx):
        if idx < 0: raise HjsonDecodeError('Expecting value', string, idx)
        try:
            return parse_object((string, idx), encoding, strict,
                _scan_once, object_hook, object_pairs_hook, memo, True)
        finally:
            memo.clear()

    return scan_once, scan_object_once


def JSONObject(state, encoding, strict, scan_once, object_hook,
        object_pairs_hook, memo=None, objectWithoutBraces=False):
    (s, end) = state
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    pairs = []

    ch, end = getNext(s, end)

    # Trivial empty object
    if not objectWithoutBraces and ch == '}':
        if object_pairs_hook is not None:
            result = object_pairs_hook(pairs)
            return result, end + 1
        pairs = {}
        if object_hook is not None:
            pairs = object_hook(pairs)
        return pairs, end + 1

    while True:
        key, end = scanKeyName(s, end, encoding, strict)
        key = memo_get(key, key)

        ch, end = getNext(s, end)
        if ch != ':':
            raise HjsonDecodeError("Expecting ':' delimiter", s, end)

        ch, end = getNext(s, end + 1)

        value, end = scan_once(s, end)
        pairs.append((key, value))

        ch, end = getNext(s, end)

        if ch == ',':
            ch, end = getNext(s, end + 1)

        if objectWithoutBraces:
            if ch == '': break;
        else:
            if ch == '}':
                end += 1
                break

        ch, end = getNext(s, end)

    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end

def JSONArray(state, scan_once):
    (s, end) = state
    values = []

    ch, end = getNext(s, end)

    # Look-ahead for trivial empty array
    if ch == ']':
        return values, end + 1
    elif ch == '':
        raise HjsonDecodeError("End of input while parsing an array (did you forget a closing ']'?)", s, end)
    _append = values.append
    while True:
        value, end = scan_once(s, end)
        _append(value)

        ch, end = getNext(s, end)
        if ch == ',':
            ch, end = getNext(s, end + 1)

        if ch == ']':
            end += 1
            break

        ch, end = getNext(s, end)

    return values, end


class HjsonDecoder(object):
    """Hjson decoder

    Performs the following translations in decoding by default:

    +---------------+-------------------+
    | JSON          | Python            |
    +===============+===================+
    | object        | dict              |
    +---------------+-------------------+
    | array         | list              |
    +---------------+-------------------+
    | string        | str, unicode      |
    +---------------+-------------------+
    | number (int)  | int, long         |
    +---------------+-------------------+
    | number (real) | float             |
    +---------------+-------------------+
    | true          | True              |
    +---------------+-------------------+
    | false         | False             |
    +---------------+-------------------+
    | null          | None              |
    +---------------+-------------------+

    """

    def __init__(self, encoding=None, object_hook=None, parse_float=None,
            parse_int=None, strict=True,
            object_pairs_hook=None):
        """
        *encoding* determines the encoding used to interpret any
        :class:`str` objects decoded by this instance (``'utf-8'`` by
        default).  It has no effect when decoding :class:`unicode` objects.

        Note that currently only encodings that are a superset of ASCII work,
        strings of other encodings should be passed in as :class:`unicode`.

        *object_hook*, if specified, will be called with the result of every
        JSON object decoded and its return value will be used in place of the
        given :class:`dict`.  This can be used to provide custom
        deserializations (e.g. to support JSON-RPC class hinting).

        *object_pairs_hook* is an optional function that will be called with
        the result of any object literal decode with an ordered list of pairs.
        The return value of *object_pairs_hook* will be used instead of the
        :class:`dict`.  This feature can be used to implement custom decoders
        that rely on the order that the key and value pairs are decoded (for
        example, :func:`collections.OrderedDict` will remember the order of
        insertion). If *object_hook* is also defined, the *object_pairs_hook*
        takes priority.

        *parse_float*, if specified, will be called with the string of every
        JSON float to be decoded.  By default, this is equivalent to
        ``float(num_str)``. This can be used to use another datatype or parser
        for JSON floats (e.g. :class:`decimal.Decimal`).

        *parse_int*, if specified, will be called with the string of every
        JSON int to be decoded.  By default, this is equivalent to
        ``int(num_str)``.  This can be used to use another datatype or parser
        for JSON integers (e.g. :class:`float`).

        *strict* controls the parser's behavior when it encounters an
        invalid control character in a string. The default setting of
        ``True`` means that unescaped control characters are parse errors, if
        ``False`` then control characters will be allowed in strings.

        """
        if encoding is None:
            encoding = DEFAULT_ENCODING
        self.encoding = encoding
        self.object_hook = object_hook
        self.object_pairs_hook = object_pairs_hook
        self.parse_float = parse_float or float
        self.parse_int = parse_int or int
        self.strict = strict
        self.parse_object = JSONObject
        self.parse_array = JSONArray
        self.parse_string = scanstring
        self.parse_mlstring = mlscanstring
        self.parse_tfnns = scantfnns
        self.memo = {}
        (self.scan_once, self.scan_object_once) = make_scanner(self)

    def decode(self, s, _PY3=PY3):
        """Return the Python representation of ``s`` (a ``str`` or ``unicode``
        instance containing a JSON document)

        """
        if _PY3 and isinstance(s, binary_type):
            s = s.decode(self.encoding)
        obj, end = self.raw_decode(s)
        ch, end = getNext(s, end)
        if end != len(s):
            raise HjsonDecodeError("Extra data", s, end, len(s))
        return obj

    def raw_decode(self, s, idx=0, _PY3=PY3):
        """Decode a JSON document from ``s`` (a ``str`` or ``unicode``
        beginning with a JSON document) and return a 2-tuple of the Python
        representation and the index in ``s`` where the document ended.
        Optionally, ``idx`` can be used to specify an offset in ``s`` where
        the JSON document begins.

        This can be used to decode a JSON document from a string that may
        have extraneous data at the end.

        """
        if idx < 0:
            # Ensure that raw_decode bails on negative indexes, the regex
            # would otherwise mask this behavior. #98
            raise HjsonDecodeError('Expecting value', s, idx)
        if _PY3 and not isinstance(s, text_type):
            raise TypeError("Input string must be text")
        # strip UTF-8 bom
        if len(s) > idx:
            ord0 = ord(s[idx])
            if ord0 == 0xfeff:
                idx += 1
            elif ord0 == 0xef and s[idx:idx + 3] == '\xef\xbb\xbf':
                idx += 3

        start_index = idx
        ch, idx = getNext(s, idx)

        # If blank or comment only file, return dict
        if start_index == 0 and ch == '':
            return {}, 0

        if ch == '{' or ch == '[':
            return self.scan_once(s, idx)
        else:
            # assume we have a root object without braces
            try:
                return self.scan_object_once(s, idx)
            except HjsonDecodeError as e:
                # test if we are dealing with a single JSON value instead (true/false/null/num/"")
                try:
                    return self.scan_once(s, idx)
                except:
                    raise e



from dbnd._core.task.data_combine_task import data_combine
from dbnd._core.utils import timezone
from dbnd._core.utils.basics.range import period_dates


__all__ = ["data_combine", "period_dates", "timezone"]



import os

import six

from targets.pipes.base import IOPipeline, NopPipeline
from targets.pipes.bzip2 import Bzip2Pipeline
from targets.pipes.gzip import GzipPipeline
from targets.pipes.text import MixedUnicodeBytesFormat, NewlinePipeline, TextPipeline


Nop = NopPipeline()
Text = TextPipeline()
UTF8 = TextPipeline(encoding="utf8")
SysNewLine = NewlinePipeline()
Gzip = GzipPipeline()
Bzip2 = Bzip2Pipeline()
MixedUnicodeBytes = MixedUnicodeBytesFormat()

SeamlessFilters = [Nop, Text]


def is_seamless_pipe(target_filter):
    return not target_filter or target_filter in SeamlessFilters



from __future__ import absolute_import

from distutils.version import LooseVersion


try:
    import marshmallow

    version_tuple = tuple(LooseVersion(marshmallow.__version__).version)
    if version_tuple != (2, 18, 0):
        raise ImportError

    from marshmallow import (
        base,
        class_registry,
        decorators,
        exceptions,
        fields,
        marshalling,
        orderedset,
        schema,
        utils,
        validate,
        warnings,
        warnings,
    )
    from marshmallow.schema import Schema, SchemaOpts, MarshalResult, UnmarshalResult
    from marshmallow.decorators import (
        pre_dump,
        post_dump,
        pre_load,
        post_load,
        validates,
        validates_schema,
    )
    from marshmallow.utils import pprint, missing
    from marshmallow.exceptions import ValidationError

except ImportError:
    from dbnd._vendor import _marshmallow as marshmallow

    from dbnd._vendor._marshmallow import (
        base,
        class_registry,
        decorators,
        exceptions,
        fields,
        marshalling,
        orderedset,
        schema,
        utils,
        validate,
        warnings,
        warnings,
    )
    from dbnd._vendor._marshmallow.schema import (
        Schema,
        SchemaOpts,
        MarshalResult,
        UnmarshalResult,
    )
    from dbnd._vendor._marshmallow.decorators import (
        pre_dump,
        post_dump,
        pre_load,
        post_load,
        validates,
        validates_schema,
    )
    from dbnd._vendor._marshmallow.utils import pprint, missing
    from dbnd._vendor._marshmallow.exceptions import ValidationError

__version__ = marshmallow.__version__
__version_info__ = marshmallow.__version_info__
__author__ = marshmallow.__author__

__all__ = [
    "Schema",
    "SchemaOpts",
    "fields",
    "validates",
    "validates_schema",
    "pre_dump",
    "post_dump",
    "pre_load",
    "post_load",
    "pprint",
    "MarshalResult",
    "UnmarshalResult",
    "ValidationError",
    "missing",
]



import copy
import logging
import sys
import typing

from typing import Callable, Dict, Type

import six

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.current import get_databand_context
from dbnd._core.errors import friendly_error
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled
from dbnd._core.utils.basics.singleton_context import SingletonContext
from dbnd._core.utils.seven import contextlib
from dbnd._vendor.snippets.luigi_registry import get_best_candidate, module_parents


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.task.task import Task
    from dbnd._core.task_build.task_definition import TaskDefinition


def _validate_no_recursion_in_config(task_name, config_task_type, param):
    if task_name == config_task_type:
        raise friendly_error.config.task_name_and_from_are_the_same(task_name, param)


class TaskEntry(object):
    def __init__(
        self, task_family, full_task_family, task_cls=None, task_cls_factory=None
    ):
        self.task_family = task_family
        self.full_task_family = full_task_family
        self._task_cls = task_cls
        self._task_cls_factory = task_cls_factory

    @property
    def task_cls(self):
        if self._task_cls is None:
            self._task_cls = self._task_cls_factory()
        return self._task_cls


class DbndTaskRegistry(SingletonContext):
    """
    Registry of all `dbnd` tasks
    """

    AMBIGUOUS_CLASS = object()

    def __init__(self):
        self._fqn_to_task_cls = (
            {}
        )  # type: Dict[str, TaskEntry]  # map between full task class name and class
        self._task_family_to_task_cls = (
            {}
        )  # type: Dict[str, TaskEntry]  # map between task_family name and class

        # namespace management

        self._namespace_map = {}

        # airflow support
        self._dag_bag = None

    def register_task(self, task_cls):
        td = task_cls.task_definition  # type: TaskDefinition
        self._register_entry(
            TaskEntry(
                task_cls=task_cls,
                task_family=td.task_family,
                full_task_family=td.full_task_family,
            )
        )

    def register_task_cls_factory(
        self, task_cls_factory, task_family, full_task_family
    ):
        # type: (Callable, str, str) -> None
        self._register_entry(
            TaskEntry(
                task_cls_factory=task_cls_factory,
                task_family=task_family,
                full_task_family=full_task_family,
            )
        )

    def _register_entry(self, task_entry):
        # type: (TaskEntry) -> None
        if task_entry.full_task_family in self._fqn_to_task_cls:
            # we're invoking factory
            existing_task_entry = self._fqn_to_task_cls[
                task_entry.full_task_family
            ]  # type: TaskEntry
            if (
                existing_task_entry._task_cls is None
                and task_entry._task_cls is not None
            ):
                existing_task_entry._task_cls = task_entry._task_cls
                return

            # else some weird duplication happened? should we log?

        self._fqn_to_task_cls[task_entry.full_task_family] = task_entry

        # TODO: support history?
        self._task_family_to_task_cls[task_entry.task_family] = (
            task_entry
            if task_entry.task_family not in self._task_family_to_task_cls
            else TaskEntry(
                task_cls=self.AMBIGUOUS_CLASS,
                task_family=task_entry.task_family,
                full_task_family=self.AMBIGUOUS_CLASS,
            )
        )

    def _get_registered_task_cls(self, name):  # type: (str) -> Type[Task]
        """
        Returns an task class based on task_family, or full task class name
        We don't preload/check anything here
        """
        task_entry = self._fqn_to_task_cls.get(name)  # type: TaskEntry
        if task_entry is None:
            task_entry = self._task_family_to_task_cls.get(name)

        if task_entry is not None:
            return task_entry.task_cls

    # used for both tasks and configurations
    def _get_task_cls(self, task_name):
        from dbnd._core.utils.basics.load_python_module import load_python_module

        task_cls = self._get_registered_task_cls(task_name)
        if task_cls:
            return task_cls

        # we are going to check if we have override/definition in config
        config_task_type = config.get(task_name, "_type", None)
        if config_task_type:
            _validate_no_recursion_in_config(task_name, config_task_type, "_type")
            try:
                return self._get_task_cls(config_task_type)
            except Exception:
                logger.error(
                    "Failed to load type required by [%s] using _type=%s",
                    task_name,
                    config_task_type,
                )
                raise
        config_task_type = config.get(task_name, "_from", None)
        if config_task_type:
            _validate_no_recursion_in_config(task_name, config_task_type, "_from")
            return self._get_task_cls(config_task_type)

        if "." in task_name:
            parts = task_name.split(".")
            possible_root_task = parts.pop()
            possible_module = ".".join(parts)

            # Try to load module and check again for existance
            load_python_module(possible_module, "task name '%s'" % task_name)

            task_cls = self._get_registered_task_cls(task_name)
            if task_cls:
                return task_cls

            # Check if task exists but user forgot to decorate method with @task
            task_module = sys.modules.get(possible_module)
            if task_module and hasattr(task_module, possible_root_task):
                user_func = getattr(task_module, possible_root_task)
                if callable(user_func):
                    from dbnd._core.tracking.python_tracking import _is_task

                    if _is_task(user_func):
                        return user_func.task_cls
                    # Non-decorated function was found - decorate and return it
                    from dbnd._core.decorator import dbnd_decorator

                    decorated_task = dbnd_decorator.task(user_func)
                    setattr(task_module, possible_root_task, decorated_task)
                    logger.warning(
                        "Found non-decorated task: %s. "
                        "Please decorate this task with the proper symbol @pipeline / @task.\n"
                        "Auto-decorating and treating it as @task ...",
                        task_name,
                    )
                    return decorated_task.task

        if is_airflow_enabled():
            from dbnd_airflow.dbnd_task_executor.airflow_operator_as_dbnd import (
                AirflowDagAsDbndTask,
            )

            dag = self._get_aiflow_dag(task_name)
            if dag:
                return AirflowDagAsDbndTask
        return None

    def get_task_cls(self, task_name):
        from dbnd._core.errors import friendly_error

        task_cls = self._get_task_cls(task_name)
        if task_cls == self.AMBIGUOUS_CLASS:
            raise friendly_error.ambiguous_task(task_name)
        if task_cls:
            return task_cls
        raise friendly_error.task_registry.task_not_exist(
            task_name=task_name,
            alternative_tasks=get_best_candidate(
                task_name, self._task_family_to_task_cls.keys()
            ),
        )

    def list_dbnd_task_classes(self):
        task_classes = []
        for task_name, task_entry in six.iteritems(self._task_family_to_task_cls):
            task_cls = task_entry.task_cls
            if task_cls == self.AMBIGUOUS_CLASS:
                continue
            td = task_cls.task_definition
            if td.hidden:
                continue
            if td.task_family.startswith("_"):
                continue

            task_classes.append(task_cls)

        task_classes = sorted(
            task_classes, key=lambda task_cls: task_cls.task_definition.full_task_family
        )
        return task_classes

    def build_dbnd_task(self, task_name, task_kwargs=None, expected_type=None):
        task_kwargs = task_kwargs or dict()
        task_kwargs.setdefault("task_name", task_name)

        task_cls = self.get_task_cls(task_name)  # type: Type[Task]
        if is_airflow_enabled():
            from dbnd_airflow.dbnd_task_executor.airflow_operator_as_dbnd import (
                AirflowDagAsDbndTask,
            )

            if issubclass(task_cls, AirflowDagAsDbndTask):
                # we are running old style dag
                dag = self._get_aiflow_dag(task_name)
                airflow_task = AirflowDagAsDbndTask.build_dbnd_task_from_dag(dag=dag)
                return airflow_task

        try:
            logger.debug("Building %s task", task_cls.task_definition.full_task_family)
            obj = task_cls(**task_kwargs)

        except Exception:
            exc = get_databand_context().settings.log.format_exception_as_str(
                sys.exc_info(), isolate=True
            )
            logger.error("Failed to build %s: \n\n%s", task_cls.get_task_family(), exc)
            raise
        if expected_type and not issubclass(task_cls, expected_type):
            raise friendly_error.task_registry.wrong_type_for_task(
                task_name, task_cls, expected_type
            )
        return obj

    def _get_aiflow_dag(self, dag_id):
        if not self._dag_bag:
            from airflow.models import DagBag

            self._dag_bag = DagBag()
        if dag_id in self._dag_bag.dags:
            return self._dag_bag.dags[dag_id]
        return None

    ########
    ## NAMESPACE MANAGEMENT
    def get_namespace(self, module_name):
        for parent in module_parents(module_name):
            #  if module is a.b.c ->  a.b can have namespace defined
            entry = self._namespace_map.get(parent)
            if entry:
                return entry
        # by default we don't have namespace
        return ""

    def register_namespace(self, scope, namespace):
        self._namespace_map[scope] = namespace


_REGISTRY = DbndTaskRegistry.try_instance()


def build_task_from_config(task_name, expected_type=None):
    tr = get_task_registry()
    return tr.build_dbnd_task(task_name=task_name, expected_type=expected_type)


def register_config_cls(config_cls):
    logger.debug("Registered config %s", config_cls)


def register_task(task):
    logger.debug("Registered task %s", task)


def get_task_registry():
    # type: ()-> DbndTaskRegistry
    return DbndTaskRegistry.get_instance()


@contextlib.contextmanager
def tmp_dbnd_registry():
    current = get_task_registry()

    new_tmp_registry = DbndTaskRegistry()
    # copy all old values
    new_tmp_registry._fqn_to_task_cls = copy.copy(current._fqn_to_task_cls)
    new_tmp_registry._task_family_to_task_cls = copy.copy(
        current._task_family_to_task_cls
    )
    new_tmp_registry._namespace_map = copy.copy(current._namespace_map)
    with DbndTaskRegistry.new_context(
        _context=new_tmp_registry, allow_override=True
    ) as r:
        # assign already known tasks
        yield r



from __future__ import absolute_import

import logging

import pandas as pd
import six

from pandas import RangeIndex
from pandas.api.types import is_categorical_dtype
from pandas.core.generic import NDFrame

from dbnd._core.errors import friendly_error
from dbnd._core.utils.structures import combine_mappings
from targets.fs import FileSystems
from targets.marshalling.marshaller import Marshaller
from targets.target_config import FileCompressions, FileFormat
from targets.utils.performance import target_timeit


logger = logging.getLogger(__name__)

PANDAS_CACHE_KEY = "pandas.DataFrame"
PANDAS_COPY_ON_READ = True

PANDAS_COMPRESSION_LIB = {
    FileCompressions.gzip: "gzip",
    FileCompressions.bzip: "bz2",
    FileCompressions.xz: "xz",
    FileCompressions.zip: "zip",
}


def _get_compression_args(target, compression_arg_name):
    file_compression = target.config.compression
    if not compression_arg_name or not file_compression:
        return {}

    compression = PANDAS_COMPRESSION_LIB.get(file_compression)
    if not compression:
        logger.info("Can't find compression lib for %s", file_compression)

    return {compression_arg_name: compression}


def _data_frame_set_index(df, set_index):
    # our frame already has index equal to the one requested by user
    if df.index.name == set_index:
        return df

    # so we have index and it's not equal to current
    if not _is_default_index(df):
        df.reset_index()

    df = df.set_index(set_index)
    return df


def _is_default_index(df):
    if df.index is None:
        return False
    return isinstance(df.index, RangeIndex) or df.index.name is None


def _file_open_mode(target, mode="r"):
    try:
        mode = mode + "b" if target.config.is_binary else mode
    except Exception as ex:
        logger.warning("Could not calculate file open mode, returning %s. %s", mode, ex)
        pass

    return mode


class _PandasMarshaller(Marshaller):
    type = pd.DataFrame
    support_directory_read = False

    _compression_read_arg = None
    _compression_write_arg = None

    support_cache = False
    disable_default_index = False

    def __init__(self, series=False):
        self.pandas_series = series

    def support_direct_read(self, target):
        # pandas actually support s3
        # TODO: decide based on pandas version and s3fs availability
        return (
            super(_PandasMarshaller, self).support_direct_access(target)
            or target.fs_name == FileSystems.s3
        )
        # s3 is supported for read, but not in write ( "_get_handler"..)

    def support_direct_write(self, target):
        # TODO: decide based on pandas version and s3fs availability
        return self.support_direct_access(target)

    def _pd_read(self, *args, **kwargs):
        pass

    def _pd_to(self, *args, **kwargs):
        pass

    @target_timeit
    def target_to_value(
        self, target, cache=True, no_copy_on_read=False, set_index=None, **kwargs
    ):
        read_kwargs = _get_compression_args(target, self._compression_read_arg).copy()
        read_kwargs.update(kwargs)

        use_cache = cache and self.support_cache
        use_cache &= target._cache.get(PANDAS_CACHE_KEY) is not None
        use_cache &= (
            "key" not in read_kwargs
        )  # support for hdf5, we don't support non default key
        if not use_cache:
            try:
                logger.info("Loading data frame from target='%s'", target)
                if self.support_direct_read(target):
                    df = self._pd_read(target.path, **read_kwargs)
                else:
                    mode = _file_open_mode(target, "r")
                    with target.open(mode) as fp:
                        df = self._pd_read(fp, **read_kwargs)
            except Exception as ex:
                raise friendly_error.failed_to_read_pandas(ex, target)
        else:
            logger.info("Loading data frame from cache: %s", target)
            df = target._cache.get(PANDAS_CACHE_KEY)
            if no_copy_on_read:
                if set_index:
                    logger.warning(
                        "You are using no_copy_on_read, with set_index, "
                        "that will change df in cache for all usages: %s" % target
                    )
            else:
                df = df.copy()

        if set_index:
            try:
                df = _data_frame_set_index(df, set_index)
            except Exception as ex:
                raise friendly_error.failed_to_set_index(ex, df, set_index, target)
        return df

    def value_to_target(self, value, target, **kwargs):
        if target.config.format != FileFormat.hdf5 and not isinstance(value, NDFrame):
            raise friendly_error.targets.failed_to_save_value__wrong_type(
                value, target, expected_type=NDFrame
            )

        # dir
        target.mkdir_parent()

        cache = kwargs.pop("cache", False)
        to_kwargs = _get_compression_args(target, self._compression_write_arg).copy()
        to_kwargs.update(kwargs)

        # if (
        #     self.file_format == StorageFormat.hdf5
        #     and to_kwargs.get("key", "data") != "data"
        # ):
        #     # support for hdf5: we can not cache non default paritions
        #     cache = False

        if cache and self.support_cache:
            if PANDAS_CACHE_KEY in target._cache is not None:
                logger.warning(
                    "Writing Pandas Data Frame to the same target for the second time! %s",
                    target,
                )
            target._cache[PANDAS_CACHE_KEY] = value

        logger.info("Saving data frame to '%s'", target)

        # we want to remove index from the formats that can't deserialize it
        # so all csv and excel should not have index saved by default
        if self.disable_default_index and _is_default_index(value):
            to_kwargs.setdefault("index", False)

        try:
            if self.support_direct_write(target):
                return self._pd_to(value, target.path, **to_kwargs)
            else:
                mode = _file_open_mode(target, "w")
                with target.open(mode) as fp:
                    return self._pd_to(value, fp, **to_kwargs)
        except Exception as ex:
            raise friendly_error.failed_to_write_pandas(ex, target)


class DataFrameToCsv(_PandasMarshaller):
    file_format = FileFormat.csv

    disable_default_index = True

    _compression_write_arg = "compression"
    _compression_read_arg = "compression"

    def _pd_read(self, *args, **kwargs):
        if self.pandas_series:
            kwargs.setdefault("squeeze", True)
            kwargs.setdefault("header", None)
        return pd.read_csv(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        return value.to_csv(*args, **kwargs)


class DataFrameToJson(_PandasMarshaller):
    file_format = FileFormat.json

    _compression_write_arg = "compression"
    _compression_read_arg = "compression"

    def _pd_read(self, *args, **kwargs):
        return pd.read_json(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        return value.to_json(*args, **kwargs)


class DataFrameToHtml(_PandasMarshaller):
    file_format = FileFormat.json

    _compression_write_arg = "compression"
    _compression_read_arg = "compression"

    def _pd_read(self, *args, **kwargs):
        return pd.read_html(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        return value.to_html(*args, **kwargs)


class DataFrameToTable(DataFrameToCsv):
    file_format = FileFormat.table


class DataFrameToPickle(_PandasMarshaller):
    file_format = FileFormat.pickle
    support_cache = True
    support_index_save = True
    _compression_write_arg = "compression"

    def _pd_read(self, *args, **kwargs):
        return pd.read_pickle(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        value.to_pickle(*args, **kwargs)


class DataFrameToParquet(_PandasMarshaller):
    file_format = FileFormat.parquet
    support_cache = True
    support_index_save = True
    support_directory_direct_read = True
    support_directory_direct_write = False

    _compression_write_arg = "compression"

    def _pd_read(self, *args, **kwargs):
        return pd.read_parquet(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        value.to_parquet(*args, **kwargs)

    def support_direct_write(self, target):
        # TODO: decide based on pandas version and s3fs availability
        return (
            super(DataFrameToParquet, self).support_direct_write(target)
            or target.fs_name == FileSystems.s3
        )


class DataFrameToFeather(_PandasMarshaller):
    support_cache = True
    support_index_save = False
    file_format = FileFormat.feather

    def _pd_read(self, *args, **kwargs):
        return pd.read_feather(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        value.to_feather(*args, **kwargs)


def _get_supported_hd5_storage_format(value):
    if (
        isinstance(value, pd.DataFrame)
        and value.select_dtypes(include=["category"]).empty
    ):
        return "fixed"

    if isinstance(value, pd.Series) and not is_categorical_dtype(value.dtypes):
        return "fixed"

    logger.warning(
        "Storage format is set to 'table' as you data contains categorical attributes. This may affect performance."
    )
    return "table"


class DataFrameToHdf5(_PandasMarshaller):
    file_format = FileFormat.hdf5
    support_cache = True
    support_index_save = True

    _compression_write_arg = "complib"

    def _pd_read(self, *args, **kwargs):
        return pd.read_hdf(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        kwargs = combine_mappings(
            {
                "key": "data",
                "format": _get_supported_hd5_storage_format(value),
                "mode": "w",
                "complib": "zlib",
            },
            kwargs,
        )
        value.to_hdf(*args, **kwargs)


class DataFrameToHdf5Table(DataFrameToHdf5):
    def _pd_to(self, value, *args, **kwargs):
        kwargs.setdefault("format", "table")
        return super(DataFrameToHdf5Table, self)._pd_to(value, *args, **kwargs)


class DataFrameDictToHdf5(_PandasMarshaller):
    file_format = FileFormat.hdf5
    support_cache = True
    support_index_save = True

    _compression_write_arg = "complib"

    def _pd_read(self, path, *args, **kwargs):
        with pd.HDFStore(path, mode="r") as store:
            return {k: store.get(k) for k in store.keys()}

    def _pd_to(self, data, file_or_path, *args, **kwargs):
        complib = kwargs.get("complib", "zlib")

        with pd.HDFStore(file_or_path, mode="w", complib=complib) as store:
            for key, partition in six.iteritems(data):
                format = kwargs.get(
                    "format", _get_supported_hd5_storage_format(partition)
                )
                store.put(key, partition, format=format, **kwargs)


class DataFrameToTsv(DataFrameToCsv):
    def _pd_read(self, path, *args, **kwargs):
        kwargs.setdefault("delimiter", "\t")
        kwargs.setdefault("error_bad_lines", False)
        return super(DataFrameToTsv, self)._pd_read(path, *args, **kwargs)

    def _pd_to(self, path, *args, **kwargs):
        kwargs.setdefault("sep", "\t")
        return super(DataFrameToTsv, self)._pd_to(path, *args, **kwargs)


class DataFrameToExcel(_PandasMarshaller):
    file_format = FileFormat.excel

    def _pd_read(self, *args, **kwargs):
        return pd.read_excel(*args, **kwargs)

    def _pd_to(self, value, *args, **kwargs):
        return value.to_excel(*args, **kwargs)



import os
import sys


def _is_init_mode():
    from subprocess import list2cmdline

    cmdline = list2cmdline(sys.argv)
    return (
        "dbnd project-init" in cmdline
        or cmdline == "-c project-init"
        or " project-init" in cmdline
        # backward compatibility
        or "dbnd init_project" in cmdline
        or cmdline == "-c init_project"
    )


def _is_running_airflow_webserver():
    # TODO: ...
    if not sys.argv or len(sys.argv) < 2:
        return False

    if sys.argv[0].endswith("airflow") and sys.argv[1] == "webserver":
        return True

    if sys.argv[0].endswith("gunicorn") and "airflow-webserver" in sys.argv:
        return True

    return False


def _init_windows_python_path(databand_package):
    # patch pwd and resource system modules
    if os.name == "nt":
        sys.path.insert(
            0,
            os.path.join(databand_package, "utils", "platform", "windows_compatible"),
        )



from __future__ import absolute_import

import logging
import os
import subprocess


logger = logging.getLogger(__name__)

GIT_ENV = "DBND_PROJECT_GIT_COMMIT"


def is_git_dirty(path=None):
    if path and os.path.isfile(path):
        path = os.path.dirname(path)
    if not os.path.exists(path):
        return True
    lines = [
        line.strip()
        for line in subprocess.check_output(
            ["git", "status", "--porcelain"], cwd=path
        ).splitlines()
        if not line.strip().startswith(b"??")
    ]

    return len(lines) > 0


def get_git_commit(path, verbose=False):
    env_commit = os.environ.get(GIT_ENV)
    if env_commit:
        return env_commit

    try:
        from git import Repo

        if os.path.isfile(path):
            path = os.path.dirname(path)

        repo = Repo(path, search_parent_directories=True)
        commit = repo.head.commit.hexsha
        return commit
    except Exception as ex:
        if verbose:
            logger.warning("Failed to get GIT version of %s: %s", path, ex)
        return None



import json
import logging
import typing

import numpy as np
import pandas as pd

from pandas.core.dtypes.common import is_bool_dtype, is_numeric_dtype, is_string_dtype


if typing.TYPE_CHECKING:
    from typing import Tuple, Dict, Optional, List
    from targets.value_meta import ValueMetaConf
    from dbnd._core.tracking.log_data_request import LogDataRequest

logger = logging.getLogger(__name__)


class PandasHistograms(object):
    """
    calculates histograms and stats on pandas dataframe.
    """

    def __init__(self, df, meta_conf):
        # type: (pd.DataFrame, ValueMetaConf) -> None
        self.df = df
        self.meta_conf = meta_conf

    def get_histograms_and_stats(self):
        # type: () -> Tuple[Dict[str, Dict], Dict[str, List[List]]]
        stats, histograms = dict(), dict()
        if self.meta_conf.log_stats:
            stats = self._calculate_stats(self.df)

        if self.meta_conf.log_histograms:
            hist_column_names = self._get_column_names_from_request(
                self.df, self.meta_conf.log_histograms
            )
            df_histograms = self.df.filter(hist_column_names)
            if pd.__version__ > "1":
                # return_type="expand" and orient="list" are used to stabilize produced
                # histograms data structure across pandas v0 & v1
                histograms = df_histograms.apply(
                    self._calculate_histograms, result_type="expand", args=(stats,)
                ).to_dict(orient="list")
            else:
                histograms = df_histograms.apply(
                    self._calculate_histograms, args=(stats,)
                ).to_dict()
                histograms = {k: list(v) for k, v in histograms.items()}

        return stats, histograms

    def _get_column_names_from_request(self, df, data_request):
        # type: (pd.DataFrame, LogDataRequest) -> List[str]
        column_names = list(data_request.include_columns)
        for column_name, column_type in df.dtypes.items():
            if data_request.include_all_string and is_string_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_boolean and is_bool_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_numeric and is_numeric_dtype(column_type):
                column_names.append(column_name)

        column_names = [
            column
            for column in column_names
            if column not in data_request.exclude_columns
        ]
        return column_names

    def _calculate_stats(self, df):
        # type: (pd.DataFrame) -> Dict[str, Dict]
        stats_column_names = self._get_column_names_from_request(
            df, self.meta_conf.log_stats
        )
        df = df.filter(stats_column_names)
        stats = df.describe(include="all").to_json()
        stats = json.loads(stats)
        stats = self._remove_none_values(stats)
        for col in stats.keys():
            stats[col]["null-count"] = np.count_nonzero(pd.isnull(df[col]))
            stats[col]["count"] = df[col].size
            stats[col]["non-null"] = stats[col]["count"] - stats[col]["null-count"]
            stats[col]["distinct"] = len(df[col].unique())
            stats[col]["type"] = self._get_column_type(df[col])
        return stats

    def _remove_none_values(self, input_dict):
        """ remove none values from dict recursively """
        for key, value in list(input_dict.items()):
            if value is None:
                input_dict.pop(key)
            elif isinstance(value, dict):
                self._remove_none_values(value)
        return input_dict

    def _get_column_type(self, column):
        # type: (pd.Series) -> str
        first_index = column.first_valid_index()
        if first_index is None:
            return column.dtype.name
        first_value = column.iat[first_index]
        return type(first_value).__name__

    def _calculate_histograms(self, df_column, stats):
        # type: (pd.Series, Dict) -> Optional[Tuple[List, List]]
        try:
            if len(df_column) == 0:
                return

            column_type = self._get_column_type(df_column)
            if is_bool_dtype(column_type) or is_string_dtype(column_type):
                counts = df_column.value_counts()  # type: pd.Series
                if (
                    stats
                    and df_column.name in stats
                    and "null-count" in stats[df_column.name]
                ):
                    null_count = stats[df_column.name]["null-count"]
                    null_column = pd.Series([null_count], index=[None])
                    counts = counts.append(null_column)
                    counts = counts.sort_values(ascending=False)
                if len(counts) > 50:
                    counts, tail = counts[:49], counts[49:]
                    tail_sum = pd.Series([tail.sum()], index=["_others"])
                    counts = counts.append(tail_sum)
                values = counts.index
            elif is_numeric_dtype(column_type):
                df_column = df_column.dropna()
                counts, values = np.histogram(
                    df_column, bins=20
                )  # type: np.array, np.array
            else:
                return

            return counts.tolist(), values.tolist()
        except Exception:
            logger.exception(
                "log_histogram: Something went wrong for column '%s'", df_column.name
            )



from dbnd.tasks.basics.sanity import dbnd_sanity_check
from dbnd.tasks.basics.shell import bash_cmd, bash_script
from dbnd.tasks.basics.simplest import SimplestPipeline, SimplestTask


try:
    import pandas
    from dbnd.tasks.basics.pandas_tasks import PandasFrameToParquet
except ImportError:
    pass



import logging
import typing

from dbnd._core.constants import (
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    MetricSource,
)
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.task_run.task_run_ctrl import TaskRunCtrl
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.timezone import utcnow
from targets import Target
from targets.values import get_value_meta_from_value


if typing.TYPE_CHECKING:
    from dbnd._core.tracking.backends import TrackingStore
    from dbnd_postgres.postgres_values import PostgresTable
    from dbnd_snowflake.snowflake_values import SnowflakeTable

    from datetime import datetime
    from typing import Any, Optional, Union, List
    import pandas as pd
    import pyspark.sql as spark

    from targets.value_meta import ValueMetaConf

logger = logging.getLogger(__name__)


class TaskRunTracker(TaskRunCtrl):
    def __init__(self, task_run, tracking_store):
        super(TaskRunTracker, self).__init__(task_run=task_run)
        self.tracking_store = tracking_store  # type: TrackingStore

    def task_run_url(self):
        run_tracker = self.run.tracker
        if not run_tracker.databand_url:
            return None

        return "{databand_url}/app/jobs/{root_task_name}/{run_uid}/{task_run_uid}".format(
            databand_url=run_tracker.databand_url,
            root_task_name=self.run.job_name,
            run_uid=self.run.run_uid,
            task_run_uid=self.task_run_uid,
        )

    # Task Handlers
    def save_task_run_log(self, log_preview, local_log_path=None):
        self.tracking_store.save_task_run_log(
            task_run=self.task_run, log_body=log_preview, local_log_path=local_log_path
        )

    def log_parameter_data(
        self, parameter, target, value, operation_type, operation_status
    ):
        # type: (TaskRunTracker, ParameterDefinition, Target, Any, DbndTargetOperationType, DbndTargetOperationStatus) -> None
        tracking_conf = self.settings.tracking
        if not tracking_conf.log_value_meta:
            return
        if value is None:
            return

        try:
            meta_conf = tracking_conf.get_value_meta_conf(
                parameter.value_meta_conf,
                value_type=parameter.value_type,
                target=target,
            )
            key = "{}.{}".format(self.task_run.task.task_name, parameter.name)
            target.target_meta = get_value_meta_from_value(key, value, meta_conf)
            # FIXME If we failed to get target meta for some reason, target operation won't be logged!
            if target.target_meta is None:
                return

            self.tracking_store.log_target(
                task_run=self.task_run,
                target=target,
                target_meta=target.target_meta,
                operation_type=operation_type,
                operation_status=operation_status,
                param_name=parameter.name,
                task_def_uid=parameter.task_definition_uid,
            )
        except Exception as ex:
            log_exception(
                "Error occurred during target logging for %s" % (target,),
                ex,
                non_critical=True,
            )

    def _log_metrics(self, metrics):
        # type: (List[Metric]) -> None
        return self.tracking_store.log_metrics(task_run=self.task_run, metrics=metrics)

    def log_artifact(self, name, artifact):
        try:
            # file storage will save file
            # db will save path
            artifact_target = self.task_run.meta_files.get_artifact_target(name)
            self.tracking_store.log_artifact(
                task_run=self.task_run,
                name=name,
                artifact=artifact,
                artifact_target=artifact_target,
            )
        except Exception as ex:
            log_exception(
                "Error occurred during log_artifact for %s" % (name,),
                ex,
                non_critical=True,
            )

    def log_metric(self, key, value, timestamp=None, source=None):
        # type: (str, Any, Optional[datetime], Optional[MetricSource]) -> None
        try:
            metric = Metric(
                key=key, value=value, source=source, timestamp=timestamp or utcnow(),
            )
            self._log_metrics([metric])
        except Exception as ex:
            log_exception(
                "Error occurred during log_metric for %s" % (key,),
                ex,
                non_critical=True,
            )

    def log_data(
        self,
        key,  # type: str
        data,  # type: Union[pd.DataFrame, spark.DataFrame, PostgresTable, SnowflakeTable]
        meta_conf,  # type: ValueMetaConf
        path=None,  # type: Optional[Union[Target,str]]
        operation_type=DbndTargetOperationType.read,  # type: DbndTargetOperationType
        operation_status=DbndTargetOperationStatus.OK,  # type: DbndTargetOperationStatus
        raise_on_error=False,  # type: bool
    ):  # type: (...) -> None
        try:
            # Combine meta_conf with the config settings
            meta_conf = self.settings.tracking.get_value_meta_conf(meta_conf)
            value_meta = get_value_meta_from_value(key, data, meta_conf, raise_on_error)
            if not value_meta:
                return

            ts = utcnow()

            if path:
                self.tracking_store.log_target(
                    task_run=self.task_run,
                    target=path,
                    target_meta=value_meta,
                    operation_type=operation_type,
                    operation_status=operation_status,
                    param_name=key,
                )
            metrics = value_meta.build_metrics_for_key(key, meta_conf)

            if metrics["user"]:
                self._log_metrics(metrics["user"])

            if metrics["histograms"]:
                self.tracking_store.log_histograms(
                    task_run=self.task_run, key=key, value_meta=value_meta, timestamp=ts
                )

            if not (metrics["user"] or metrics["histograms"] or path):
                logger.info("No metrics to log_data(key={}, data={})".format(key, data))

        except Exception as ex:
            log_exception(
                "Error occurred during log_dataframe for %s" % (key,),
                ex,
                non_critical=not raise_on_error,
            )
            if raise_on_error:
                raise



# -*- coding: utf-8 -*-
#
# Copyright 2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import locale
import os
import re
import warnings

import six

from targets.pipes.base import BaseWrapper, WrappedFormat


class NewlineWrapper(BaseWrapper):
    def __init__(self, stream, newline=None):
        if newline is None:
            self.newline = newline
        else:
            self.newline = newline.encode("ascii")

        if self.newline not in (b"", b"\r\n", b"\n", b"\r", None):
            raise ValueError(
                "newline need to be one of {b'', b'\r\n', b'\n', b'\r', None}"
            )
        super(NewlineWrapper, self).__init__(stream)

    def read(self, n=-1):
        b = self._stream.read(n)

        if self.newline == b"":
            return b

        if self.newline is None:
            newline = b"\n"

        return re.sub(b"(\n|\r\n|\r)", newline, b)

    def writelines(self, lines):
        if self.newline is None or self.newline == "":
            newline = os.linesep.encode("ascii")
        else:
            newline = self.newline

        self._stream.writelines(
            (re.sub(b"(\n|\r\n|\r)", newline, line) for line in lines)
        )

    def write(self, b):
        if self.newline is None or self.newline == "":
            newline = os.linesep.encode("ascii")
        else:
            newline = self.newline

        self._stream.write(re.sub(b"(\n|\r\n|\r)", newline, b))


class TextWrapper(io.TextIOWrapper):
    def __exit__(self, *args):
        # io.TextIOWrapper close the file on __exit__, let the underlying file decide
        if not self.closed and self.writable():
            super(TextWrapper, self).flush()

        self._stream.__exit__(*args)

    def __del__(self, *args):
        # io.TextIOWrapper close the file on __del__, let the underlying file decide
        if not self.closed and self.writable():
            super(TextWrapper, self).flush()

        try:
            self._stream.__del__(*args)
        except AttributeError:
            pass

    def __init__(self, stream, *args, **kwargs):
        self._stream = stream
        try:
            super(TextWrapper, self).__init__(stream, *args, **kwargs)
        except TypeError:
            pass

    def __getattr__(self, name):
        if name == "_stream":
            raise AttributeError(name)
        return getattr(self._stream, name)

    def __enter__(self):
        self._stream.__enter__()
        return self


class MixedUnicodeBytesWrapper(BaseWrapper):
    """
    """

    def __init__(self, stream, encoding=None):
        if encoding is None:
            encoding = locale.getpreferredencoding()
        self.encoding = encoding
        super(MixedUnicodeBytesWrapper, self).__init__(stream)

    def write(self, b):
        self._stream.write(self._convert(b))

    def writelines(self, lines):
        self._stream.writelines((self._convert(line) for line in lines))

    def _convert(self, b):
        if isinstance(b, six.text_type):
            b = b.encode(self.encoding)
            warnings.warn("Writing unicode to byte stream", stacklevel=2)
        return b


class TextPipeline(WrappedFormat):

    input = "unicode"
    output = "bytes"
    wrapper_cls = TextWrapper


class MixedUnicodeBytesFormat(WrappedFormat):

    output = "bytes"
    wrapper_cls = MixedUnicodeBytesWrapper


class NewlinePipeline(WrappedFormat):

    input = "bytes"
    output = "bytes"
    wrapper_cls = NewlineWrapper



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides a class :class:`MockTarget`, an implementation of :py:class:`~databand.targets.Target`.
:class:`MockTarget` contains all data in-memory.
The main purpose is unit testing workflows without writing to disk.
"""

import multiprocessing
import sys

from io import BytesIO

import six

from targets import target
from targets.fs.file_system import FileSystem


class MockFileSystem(FileSystem):
    """
    MockFileSystem inspects/modifies _data to simulate file system operations.
    """

    _data = None
    instance = None

    def __init__(self, mirror_on_stderr=False):
        super(MockFileSystem, self).__init__()

        self._mirror_on_stderr = mirror_on_stderr

    def copy(self, path, dest, raise_if_exists=False):
        """
        Copies the contents of a single file path to dest
        """
        if raise_if_exists and dest in self.get_all_data():
            raise RuntimeError("Destination exists: %s" % path)
        contents = self.get_all_data()[path]
        self.get_all_data()[dest] = contents

    def get_all_data(self):
        # This starts a server in the background, so we don't want to do it in the global scope
        if MockFileSystem._data is None:
            MockFileSystem._data = multiprocessing.Manager().dict()
        return MockFileSystem._data

    def get_data(self, fn):
        return self.get_all_data()[fn]

    def exists(self, path):
        return path in self.get_all_data()

    def remove(self, path, recursive=True, skip_trash=True):
        """
        Removes the given mockfile. skip_trash doesn't have any meaning.
        """
        if recursive:
            to_delete = []
            for s in self.get_all_data().keys():
                if s.startswith(path):
                    to_delete.append(s)
            for s in to_delete:
                self.get_all_data().pop(s)
        else:
            self.get_all_data().pop(path)

    def move(self, path, dest, raise_if_exists=False):
        """
        Moves a single file from path to dest
        """
        if raise_if_exists and dest in self.get_all_data():
            raise RuntimeError("Destination exists: %s" % path)
        contents = self.get_all_data().pop(path)
        self.get_all_data()[dest] = contents

    def listdir(self, path):
        """
        listdir does a prefix match of self.get_all_data(), but doesn't yet support globs.
        """
        return [s for s in self.get_all_data().keys() if s.startswith(path)]

    def isdir(self, path):
        return any(self.listdir(path))

    def mkdir(self, path, parents=True, raise_if_exists=False):
        """
        mkdir is a noop.
        """
        pass

    def clear(self):
        self.get_all_data().clear()

    def open_write(self, path, mode="w", **kwargs):
        return MockFile(path=path, fs=self, mode=mode)

    def open_read(self, path, mode="r"):
        raise NotImplementedError()
        # return StringBuffer(self.get_all_data()[path])


class MockFile(BytesIO):
    # Just to be able to do writing + reading from the same buffer

    _write_line = True

    def __init__(self, path, fs, mode):
        self.path = path
        self.fs = fs
        self.mode = mode

    def set_wrapper(self, wrapper):
        self.wrapper = wrapper

    def write(self, data):
        if self.fs._mirror_on_stderr:
            if self._write_line:
                sys.stderr.write(self.path + ": ")
            if six.binary_type:
                sys.stderr.write(data.decode("utf8"))
            else:
                sys.stderr.write(data)
            if (data[-1]) == "\n":
                self._write_line = True
            else:
                self._write_line = False
        super(MockFile, self).write(data)

    def close(self):
        if self.mode == "w":
            # try:
            #     self.target.wrapper.flush()
            # except AttributeError:
            #     pass
            self.fs.get_all_data()[self.path] = self.getvalue()
        super(MockFile, self).close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            self.close()

    def __enter__(self):
        return self

    def readable(self):
        return self.mode == "r"

    def writeable(self):
        return self.mode == "w"

    def seekable(self):
        return False


MockFileSystem.instance = MockFileSystem()


def mock_target(*path, **kwargs):
    return target(*path, fs=MockFileSystem.instance, **kwargs)



import abc
import logging
import os.path
import random
import tempfile

from dbnd._core.current import try_get_databand_context, try_get_databand_run


logger = logging.getLogger(__name__)

_CONFIG_PARSER = None


class TargetConfigProvider(object):
    @abc.abstractmethod
    def get_config_section_values(self):
        pass

    @abc.abstractmethod
    def get_config_value(self, section, value):
        pass


def set_config_provider(config_parser):
    global _CONFIG_PARSER
    _CONFIG_PARSER = config_parser


def get_config_section_values(section):
    config_provider = _CONFIG_PARSER
    if not config_provider:
        return {}
    return config_provider.get_config_section_values(section)


def get_local_tempfile(*path):
    run = try_get_databand_run()
    if run:
        dbnd_local_root = run.get_current_dbnd_local_root()
        if dbnd_local_root.exists():
            # on remote engine temp defined at driver can be un-awailable
            # simple workaround to use tmp folder on the machine
            tempdir = dbnd_local_root.partition("tmp").path
        else:
            # fallback to simple temp dir
            tempdir = tempfile.gettempdir()
    else:
        tempdir = tempfile.gettempdir()

    path = os.path.join(tempdir, "databand-tmp-%09d" % random.randrange(0, 1e10), *path)
    base_dir = os.path.dirname(path)
    try:
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
    except Exception as ex:
        logger.info("Failed to create temp dir %s: %s", base_dir, ex)
    return path


def is_in_memory_cache_target_value():
    dc = try_get_databand_context()
    if dc:
        return dc.settings.features.in_memory_cache_target_value
    return False



import cProfile
import logging
import os

from decorator import contextmanager


@contextmanager
def perf_trace(file):
    pr = cProfile.Profile()
    pr.enable()
    yield pr
    pr.disable()

    file = os.path.abspath(file)
    if not os.path.exists(os.path.dirname(file)):
        os.makedirs(os.path.dirname(file))

    pr.dump_stats(file)
    logging.warning("Performance report saved at %s", file)



from dbnd._core.errors import DatabandConfigError, TaskClassNotFoundException


def wrong_type_for_task(section, task_cls, expected_type):
    return DatabandConfigError(
        "Your config '{section}' should be derived from '{expected_type}. Got {task_cls}".format(
            section=section, expected_type=expected_type, task_cls=task_cls
        ),
        help_msg="Make sure '{section}' section in your configuration is of the correct type! (check [{"
        "section}]._type and [{section}]._from ".format(
            section=section,
            task_cls=task_cls,
            # expected_type=expected_type,
        ),
    )


def task_not_exist(task_name, alternative_tasks=None, module=None):
    err_msg = "Could not find the requested task/function '%s'" % task_name
    if module:
        err_msg += " in the current module %s." % module
    if alternative_tasks:
        err_msg += "\nDid you mean to call one of the following:\n %s" % "\n ".join(
            alternative_tasks
        )

    err_msg += (
        "\nPlease check the requested name, make sure the correct module installed, validate the value or add [%s]_type=EXPECTEDTYPE to configuration"
        % task_name
    )

    return TaskClassNotFoundException(
        err_msg,
        help_msg="Validate that this method exist in the requested module. "
        "Make sure you typed the function name correctly.",
        show_exc_info=False,
    )



import itertools
import json
import logging
import typing

from ast import literal_eval
from collections import OrderedDict
from typing import Iterable

import six

from dbnd._core.errors import DatabandConfigError
from dbnd._core.utils import json_utils
from dbnd._core.utils.traversing import traverse
from dbnd._vendor.splitter import split_args, unquote
from targets.values.value_type import ValueType


logger = logging.getLogger(__name__)

_PARSABLE_PARAM_PREFIX = u"@{["


#
# if issubclass(obj_type, (typing.List, list)) or obj_type == typing.Sequence:
#     return self._structured_type(obj_type, ListValueType)
# if issubclass(obj_type, (typing.Dict, dict)):
#     return self._structured_type(obj_type, DictValueType)
# if issubclass(obj_type, (typing.Tuple, tuple)):
#     return self._structured_type(obj_type, TupleValueType)


class _StructureValueType(ValueType):
    sub_value_type = None

    def __init__(self, sub_value_type=None):
        super(_StructureValueType, self).__init__()
        self.sub_value_type = sub_value_type or self.__class__.sub_value_type

    def is_type_of(self, value):
        return isinstance(value, self.type)

    def is_handler_of_type(self, type_):
        return issubclass(type_, self.type)

    def parse_from_str(self, x):
        """
               Parses an immutable and ordered ``dict`` from a JSON string using standard JSON library.
        Parse an individual value from the input.

        """

        # if isinstance(value, Mapping):
        #     # we are good to go, it'x dictionary already
        #     return value
        if not x:
            return self._generate_empty_default()

        # this is string and we need to parse it
        if not isinstance(x, six.string_types):
            raise DatabandConfigError(
                "Can't parse '%x' into parameter. Value should be string" % x
            )

        x = x.strip()
        if not x:
            return self._generate_empty_default()

        if x[0] in _PARSABLE_PARAM_PREFIX:
            value = json_utils.loads(x)
        else:
            value = self._parse_from_str_simple(x)

            if not self.is_type_of(value):
                raise DatabandConfigError(
                    "Can't parse '%s' into %s" % (value, self.type)
                )
        if self.sub_value_type:
            value = traverse(value, self.sub_value_type.parse_value)

        return value

    def parse_from_str_lines(self, lines):
        value = lines
        if self.sub_value_type:
            value = traverse(value, self.sub_value_type.parse_from_str)
        return value

    def normalize(self, value):
        if self.sub_value_type:
            value = traverse(value, self.sub_value_type.normalize)
        return value

    def parse_value(self, value, load_value=None, target_config=None):
        """
        parse structure first
        parse every element
        """
        if value is None:
            return value

        if isinstance(value, six.string_types):
            return super(_StructureValueType, self).parse_value(
                value=value, load_value=load_value, target_config=target_config
            )
        else:
            if self.sub_value_type:
                value = traverse(
                    struct=value, convert_f=self.sub_value_type.parse_value
                )
        return value

    def load_runtime(self, value, **kwargs):
        if self.sub_value_type:
            return traverse(value, self.sub_value_type.load_runtime)
        return value

    def _parse_from_str_simple(self, str_value):
        raise NotImplementedError("non parsable inputs are not supported for %s" % self)

    def to_str(self, x):
        if self.sub_value_type:
            x = traverse(x, self.sub_value_type.to_str)
        return json_utils.dumps_safe(x)

    def to_str_lines(self, x):
        if self.sub_value_type:
            x = traverse(x, self.sub_value_type.to_str)
        return x

    def with_sub_type_handler(self, type_handler):
        # bug in python <3.7  (copy of generic)
        # new_self = copy.copy(self)
        # new_self.sub_value_type = type_handler
        # return new_self
        return self.__class__(sub_value_type=type_handler)

    def __iter__(self):
        raise Exception("Do not iterate, this one is for type hinting only")

    def __repr__(self):
        sub_value_type_repr = ""
        if self.sub_value_type:
            sub_value_type_repr = "[{}]".format(repr(self.sub_value_type))
        return "{self.type_str}{sub_value_type_repr}@{self.__class__.__name__}".format(
            self=self, sub_value_type_repr=sub_value_type_repr
        )

    def __str__(self):
        sub_value_type_repr = ""
        if self.sub_value_type:
            sub_value_type_repr = "[{}]".format(str(self.sub_value_type))
        return "{self.type_str}{sub_value_type_repr}".format(
            self=self, sub_value_type_repr=sub_value_type_repr
        )


class DictValueType(_StructureValueType):
    """
    Parameter whose value is a ``dict``.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
            tags = databand.DictParameter()

            def run(self):
                logging.info("Find server with role: %s", self.tags['role'])
                server = aws.ec2.find_my_resource(self.tags)


    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --tags <JSON string>
        $ dbnd --module my_tasks MyTask --tags 'key=value key="some value"'

    Simple example with two tags:

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --tags '{"role": "web", "env": "staging"}'

    We also enable to provide multiple values per parameter

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --tags '{"role": "web"}'  --tags  '{ "env": "staging"}'


    It can be used to define dynamic parameters, when you do not know the exact list of your parameters (e.g. list of
    tags, that are dynamically constructed outside databand), or you have a complex parameter containing logically related
    values (like a database connection config).
    """

    type = typing.Dict

    type_str_extras = ("dict",)

    def _generate_empty_default(self):
        return dict()

    def _parse_from_str_simple(self, str_value):
        p_value = {}
        for param in split_args(str_value):
            if "=" not in param:
                raise ValueError(
                    "Failed to parse '%s', expected format is KEY=VALUE", param
                )
            name, sep, var = param.partition("=")
            p_value[name.strip()] = unquote(var.strip())
        return p_value

    def __str__(self):
        sub_value_type_repr = ""
        if self.sub_value_type:
            sub_value_type_repr = "[Any,{}]".format(str(self.sub_value_type))
        return "{self.type_str}{sub_value_type_repr}".format(
            self=self, sub_value_type_repr=sub_value_type_repr
        )


class ListValueType(_StructureValueType, Iterable):
    """
    Value whose value is a ``list``.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
          grades = databand.ListParameter()

            def run(self):
                sum = 0
                for element in self.grades:
                    sum += element
                avg = sum / len(self.grades)


    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --grades <JSON string>
        $ dbnd --module my_tasks MyTask --grades '[100,70]'

    """

    type = typing.List
    type_str_extras = ("list", "DataList")

    config_name = "list"
    support_merge = True  # we know how to merge frames

    def _generate_empty_default(self):
        return list()

    def _parse_from_str_simple(self, value):
        return value.split(",")

    def merge_values(self, *values):
        return list(itertools.chain(*values))


class SetValueType(_StructureValueType):
    """
    Value whose value is a ``set``.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
          grades = databand.ListParameter()

            def run(self):
                sum = 0
                for element in self.grades:
                    sum += element
                avg = sum / len(self.grades)


    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --grades <JSON string>

    Simple example with two grades:

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --grades '[100,70]'
    """

    type = typing.Set

    def _generate_empty_default(self):
        return set()

    def _parse_from_str_simple(self, value):
        return set(value.split(","))

    def normalize(self, value):
        if isinstance(value, set):
            return value
        return set(value)

    def to_str(self, x):
        if self.sub_value_type:
            x = traverse(x, self.sub_value_type.to_str)

        # we sort the set before we serialize!
        x = sorted(x, key=lambda x: json_utils.dumps_canonical(x))
        return json_utils.dumps_canonical(x)


class TupleValueType(_StructureValueType):
    """
    Parameter whose value is a ``tuple`` or ``tuple`` of tuples.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
          book_locations = databand.TupleParameter()

            def run(self):
                for location in self.book_locations:
                    print("Go to page %d, line %d" % (location[0], location[1]))


    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --book_locations <JSON string>

    Simple example with two grades:

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --book_locations '((12,3),(4,15),(52,1))'
    """

    type = typing.Tuple

    type_str_extras = ("tuple",)

    config_name = "tuple"

    def _parse_from_str_simple(self, value):
        return value.split(",")

    def _generate_empty_default(self):
        return tuple()

    def parse_from_str(self, x):
        """
        Parse an individual value from the input.

        :param str x: the value to parse.
        :return: the parsed value.
        """
        # Since the result of json.dumps(tuple) differs from a tuple string, we must handle either case.
        # A tuple string may come from a config file or from cli execution.

        # t = ((1, 2), (3, 4))
        # t_str = '((1,2),(3,4))'
        # t_json_str = json.dumps(t)
        # t_json_str == '[[1, 2], [3, 4]]'
        # json.loads(t_json_str) == t
        # json.loads(t_str) == ValueError: No JSON object could be decoded

        # Therefore, if json.loads(x) returns a ValueError, try ast.literal_eval(x).
        # ast.literal_eval(t_str) == t
        # hjson will not handle that!
        try:
            # loop required to parse tuple of tuples
            return tuple(tuple(x) for x in json.loads(x, object_pairs_hook=OrderedDict))
        except ValueError:
            return literal_eval(x)  # if this causes an error, let that error be raised.

    def is_handler_of_type(self, type_):
        return type_ in [typing.Tuple, tuple]



from typing import Any, Optional

import attr


@attr.s(slots=True)
class ConfigValue(object):
    """
    Represent config value in config
    """

    value = attr.ib()  # type: Any

    # only value is important, everything else should be cmd=False
    # source = attr.ib(eq=False)  # type: Optional[str]
    # require_parse = attr.ib(default=True, repr=False, eq=False)  # type: bool
    # override = attr.ib(default=False, repr=False, eq=False)  # type: bool
    # set_if_not_exists_only = attr.ib(default=False, repr=False, eq=False)  # type: bool
    #
    # deprecate_message = attr.ib(default=None, repr=False, eq=False)
    # warnings = attr.ib(factory=list, repr=False, eq=False)

    # Airflow monitor does not support 'eq' in attr.ib()
    source = attr.ib(default=False)  # type: Optional[str]
    require_parse = attr.ib(default=True)  # type: bool
    override = attr.ib(default=False)  # type: bool
    set_if_not_exists_only = attr.ib(default=False)  # type: bool

    deprecate_message = attr.ib(default=None)
    warnings = attr.ib(factory=list)



from dbnd._vendor import click


@click.command()
def ipython():
    """Get ipython shell with Databand's context"""
    # noinspection PyUnresolvedReferences
    from dbnd_web import models  # noqa
    from dbnd import new_dbnd_context
    from airflow.utils.db import create_session
    import IPython

    with new_dbnd_context(
        name="ipython", autoload_modules=False
    ) as ctx, create_session() as session:
        header = "\n\t".join(
            [
                "Welcome to \033[91mDataband\033[0m's ipython command.\nPredefined variable are",
                "\033[92m\033[1mctx\033[0m     -> dbnd_context",
                "\033[92m\033[1msession\033[0m -> DB session",
                "\033[92m\033[1mmodels\033[0m  -> dbnd models",
            ]
        )
        IPython.embed(colors="neutral", header=header)



import logging
import os

from collections import defaultdict, namedtuple
from datetime import datetime

import yaml

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.utils.string_utils import pluralize
from dbnd._core.utils.timezone import make_aware
from dbnd.api.scheduler import (
    get_scheduled_jobs,
    patch_scheduled_job,
    post_scheduled_job,
)
from dbnd.api.shared_schemas.scheduled_job_schema import (
    ScheduledJobSchemaV2,
    validate_cron,
)


logger = logging.getLogger(__name__)


class InvalidConfigException(Exception):
    pass


DeltaResult = namedtuple("DeltaResult", "to_create to_update to_enable to_disable")

# this horrible hack is because otherwise the yaml loader will decide to be "helpful" and convert anything that looks like a datetime
# to a datetime object. This later collides with marshmallow trying to parse and validate what it expects to be a datetime strings
NoDatesSafeLoader = yaml.SafeLoader
NoDatesSafeLoader.yaml_implicit_resolvers = {
    k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
    for k, v in NoDatesSafeLoader.yaml_implicit_resolvers.items()
}


class SchedulerFileConfigLoader(object):
    def __init__(self, config_file=None):
        config.load_system_configs()
        self.config_file = (
            config_file if config_file else config.get("scheduler", "config_file")
        )
        self.active_by_default = config.get("scheduler", "active_by_default")

    def sync(self):
        if not self.config_file:
            return

        to_create, to_update, to_enable, to_disable = self.build_delta()

        for s in to_create:
            post_scheduled_job(s)
        for s in to_update + to_disable + to_enable:
            patch_scheduled_job(s)

        if to_create or to_update or to_enable or to_disable:
            return {
                "created": len(to_create),
                "updated": len(to_update),
                "enabled (previously deleted)": len(to_enable),
                "disabled (deleted from file)": len(to_disable),
            }

    def build_delta(self):
        from_file = self.read_config(self.config_file)

        file_modified_time = make_aware(
            datetime.utcfromtimestamp(os.path.getmtime(self.config_file))
        )
        if not from_file:
            from_file = []

        from_file = self.load_and_validate(from_file)

        from_db = [
            s["DbndScheduledJob"]
            for s in get_scheduled_jobs(from_file_only=True, include_deleted=None)
        ]
        return self.delta(from_db, from_file, file_modified_time)

    def read_config(self, config_file):
        if not config_file:
            return []

        with open(config_file, "r") as stream:
            try:
                config_entries = yaml.load(stream, Loader=NoDatesSafeLoader)
                if not config_entries:
                    return []

                if not isinstance(config_entries, list):
                    raise InvalidConfigException(
                        "failed to load scheduler config at %s: invalid yaml format. Expecting a list of objects. Parsed type: %s"
                        % (config_file, config_entries.__class__.__name__)
                    )

                for i, s in enumerate(config_entries):
                    s["list_order"] = i
                return config_entries
            except yaml.YAMLError as exc:
                raise InvalidConfigException(
                    "failed to load scheduler config at %s: %s" % (config_file, exc)
                )

    def load_and_validate(self, file_entries):
        schema = ScheduledJobSchemaV2(strict=False)

        # validate entries against the schema and check the schedule_interval
        loaded_entries = []
        for file_entry in file_entries:
            loaded_entry, errors = schema.load(file_entry)
            loaded_entry = loaded_entry["DbndScheduledJob"]
            loaded_entries.append(loaded_entry)
            if errors:
                clean_errors = []
                for field, field_errors in errors.items():
                    clean_errors.append("%s: %s" % (field, ", ".join(field_errors)))
                loaded_entry["validation_errors"] = clean_errors

            loaded_entry["validation_errors"] = (
                loaded_entry["validation_errors"]
                if "validation_errors" in loaded_entry
                else []
            )

            if "schedule_interval" in loaded_entry:
                validation_error = validate_cron(loaded_entry["schedule_interval"])
                if validation_error:
                    loaded_entry["validation_errors"].append(
                        "Invalid schedule_interval: %s" % validation_error
                    )

        # check for duplicates
        group_by_name = defaultdict(lambda: [])
        for s in loaded_entries:
            if "name" in s:
                group_by_name[s["name"]].append(s)
        duplicates = [entries for entries in group_by_name.values() if len(entries) > 1]
        for duplicate_set in duplicates:
            for entry in duplicate_set:
                entry["validation_errors"].append(
                    "%s other %s exist in the configuration file with the same name"
                    % (
                        len(duplicate_set) - 1,
                        pluralize("entry", len(duplicate_set) - 1, "entries"),
                    )
                )

        # format and log validation errors
        log_message = ""
        for i, config_entry in enumerate(loaded_entries):
            if config_entry["validation_errors"]:
                config_entry["validation_errors"] = "\t" + "\n\t".join(
                    config_entry["validation_errors"]
                )
                log_message += "\n%s:\n%s" % (
                    config_entry["name"]
                    if "name" in config_entry and config_entry["name"]
                    else "<entry #%s name missing>" % i,
                    config_entry["validation_errors"],
                )
            else:
                config_entry["validation_errors"] = ""

        if log_message:
            logger.error("scheduler config file validation errors:%s" % log_message)

        # name is the key, so we can't save to the db without the key
        for entry in loaded_entries:
            if "name" not in entry:
                loaded_entries.remove(entry)

        return loaded_entries

    def delta(self, from_db, from_file, file_modified_time):
        from_db_by_name = {s["name"]: s for s in from_db}
        from_file_by_name = {s["name"]: s for s in from_file}

        to_create = []
        to_update = []
        to_enable = []
        to_disable = []

        for s in from_file:
            if s["name"] in from_db_by_name:
                db_s = from_db_by_name[s["name"]]
                if db_s.get("deleted_from_file", False):
                    s["deleted_from_file"] = False
                    s["active"] = s.get("active", self.active_by_default)
                    to_enable.append(s)
                elif self._key_diff(s, db_s):
                    # if we don't do this then if the user changes the active state from the ui
                    # it will constantly be overriden by the active state from the file
                    if (
                        db_s.get("update_time", None)
                        and db_s["update_time"] > file_modified_time
                        and "active" in s
                    ):
                        del s["active"]

                    if self._key_diff(s, db_s):
                        to_update.append(s)
            else:
                s["from_file"] = True
                s["active"] = s["active"] if "active" in s else self.active_by_default
                to_create.append(s)

        for db_s in from_db:
            if db_s["name"] not in from_file_by_name and not db_s.get(
                "deleted_from_file", False
            ):
                to_disable.append(
                    {"name": db_s["name"], "deleted_from_file": True, "active": False}
                )

        return DeltaResult(to_create, to_update, to_enable, to_disable)

    def _key_diff(self, a, b):
        return any((key for key in a if key not in b or a[key] != b[key]))



from __future__ import absolute_import

import logging
import os
import time
import typing

from typing import Dict

import pandas as pd
import six

from pandas.core.util.hashing import hash_pandas_object

from dbnd._core.errors import friendly_error
from dbnd._vendor import fast_hasher
from targets.target_config import FileFormat
from targets.value_meta import ValueMeta
from targets.values.builtins_values import DataValueType
from targets.values.pandas_histograms import PandasHistograms
from targets.values.structure import DictValueType
from targets.values.value_type import _isinstances


if typing.TYPE_CHECKING:
    from targets.value_meta import ValueMetaConf

logger = logging.getLogger(__name__)


class DataFrameValueType(DataValueType):
    type = pd.DataFrame
    type_str = "DataFrame"
    support_merge = True

    config_name = "pandas_dataframe"

    def to_signature(self, x):
        shape = "[%s]" % (",".join(map(str, x.shape)))
        return "%s:%s" % (shape, fast_hasher.hash(x))

    def to_preview(self, df, preview_size):  # type: (pd.DataFrame, int) -> str
        return df.to_string(index=False, max_rows=20, max_cols=1000)[:preview_size]

    def get_value_meta(self, value, meta_conf):
        # type: (pd.DataFrame, ValueMetaConf) -> ValueMeta
        data_schema = {}
        if meta_conf.log_schema:
            data_schema.update(
                {
                    "type": self.type_str,
                    "columns": list(value.columns),
                    "shape": value.shape,
                    "dtypes": {col: str(type_) for col, type_ in value.dtypes.items()},
                }
            )

        if meta_conf.log_size:
            data_schema["size"] = int(value.size)

        value_preview, data_hash = None, None
        if meta_conf.log_preview:
            value_preview = self.to_preview(
                value, preview_size=meta_conf.get_preview_size()
            )
            try:
                data_hash = fast_hasher.hash(
                    hash_pandas_object(value, index=True).values
                )
            except Exception as e:
                logger.warning(
                    "Could not hash dataframe object %s! Exception: %s", value, e
                )

        if meta_conf.log_histograms:
            start_time = time.time()
            stats, histograms = PandasHistograms(
                value, meta_conf
            ).get_histograms_and_stats()
            hist_sys_metrics = {
                "histograms_and_stats_calc_time": time.time() - start_time
            }
        else:
            stats, histograms = {}, {}
            hist_sys_metrics = None

        return ValueMeta(
            value_preview=value_preview,
            data_dimensions=value.shape,
            data_schema=data_schema,
            data_hash=data_hash,
            descriptive_stats=stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
        )

    def merge_values(self, *values, **kwargs):
        # Concatenate all data into one DataFrame
        # We don't want list to be stored in memory
        # however, concat does list() on the iterator as one of the first things

        # # right now we can base our decision on:
        # # if our parts have RangeIndex (_default_index())
        # # let ignore it and have new index

        # we know that some formats can't have an index,
        # but let just base our decision on RangeIndex
        # file_format = kwargs.get("file_format", None)
        # storage_format_support_index = False
        # if file_format:
        #     storage_format_support_index = file_format in PANDAS_CACHE_SUPPORTED_FORMATS
        # else:
        #     all_formats = {t.file_format for t in targets}
        #     storage_format_support_index = all(f in PANDAS_CACHE_SUPPORTED_FORMATS for f in all_formats)

        meaningful_index = kwargs.get("set_index")
        # if we don't have index based on some data column
        if not meaningful_index:
            from targets.marshalling.pandas import _is_default_index

            meaningful_index = all(not _is_default_index(df) for df in values)

        return pd.concat(
            values, verify_integrity=True, ignore_index=not meaningful_index
        )


class PandasSeriesValueType(DataFrameValueType):
    type = pd.Series
    type_str = "Series"
    support_merge = False

    config_name = "pandas_series"


class DataFramesDictValueType(DictValueType):
    type = Dict[str, pd.DataFrame]
    type_str = "Dict[str,DataFrame]"
    type_str_extras = ("DataFrameDict", "Dict[str,pd.DataFrame]")
    config_name = "pandas_dataframe"
    sub_value_type = DataFrameValueType()

    def is_handler_of_type(self, type_):
        return type_ == self.type

    def is_type_of(self, value):
        if value is None:
            return False
        if not isinstance(value, dict):
            return False
        return value and _isinstances(value.values(), pd.DataFrame)

    def load_from_target(self, target, **kwargs):
        if target.config.format == FileFormat.hdf5:
            return target.load(value_type=Dict[str, pd.DataFrame])

        from targets import DirTarget

        if isinstance(target, DirTarget):

            result = {}
            for p in target.list_partitions():
                p_name = "/" + os.path.splitext(os.path.basename(p.path))[0]
                result[p_name] = p.load(value_type=pd.DataFrame, **kwargs)
            return result
        else:
            return target.load(value_type=Dict, **kwargs)

    def save_to_target(self, target, value, **kwargs):
        from targets.dir_target import DirTarget

        if not isinstance(value, dict):
            raise friendly_error.targets.data_frame_dict_value_should_be_dict(
                target, value
            )
        if isinstance(target, DirTarget):
            for key, partition_value in six.iteritems(value):
                target.partition(key).dump(
                    value=partition_value, value_type=pd.DataFrame
                )
            target.mark_success()
        else:
            # if target.config.format == FileFormat.hdf5:
            target.dump(value=value, value_type=self)

    def __str__(self):
        return "Dict[Any,DataFrame]"



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import os
import signal
import subprocess

from targets.config import get_local_tempfile


class FileWrapper(object):
    """
    Wrap `file` in a "real" so stuff can be added to it after creation.
    """

    def __init__(self, file_object):
        self._subpipe = file_object

    def __getattr__(self, name):
        # forward calls to 'write', 'close' and other methods not defined below
        return getattr(self._subpipe, name)

    def __enter__(self, *args, **kwargs):
        # instead of returning whatever is returned by __enter__ on the subpipe
        # this returns self, so whatever custom injected methods are still available
        # this might cause problems with custom file_objects, but seems to work
        # fine with standard python `file` objects which is the only default use
        return self

    def __exit__(self, *args, **kwargs):
        return self._subpipe.__exit__(*args, **kwargs)

    def __iter__(self):
        return iter(self._subpipe)


class InputPipeProcessWrapper(object):
    def __init__(self, command, input_pipe=None):
        """
        Initializes a InputPipeProcessWrapper instance.

        :param command: a subprocess.Popen instance with stdin=input_pipe and
                        stdout=subprocess.PIPE.
                        Alternatively, just its args argument as a convenience.
        """
        self._command = command

        self._input_pipe = input_pipe
        self._original_input = True

        if input_pipe is not None:
            try:
                input_pipe.fileno()
            except AttributeError:
                # subprocess require a fileno to work, if not present we copy to disk first
                self._original_input = False
                self._tmp_file = get_local_tempfile("databand-process_tmp")
                with open(self._tmp_file, "wb") as f:
                    self._tmp_file = f.name
                    while True:
                        chunk = input_pipe.read(io.DEFAULT_BUFFER_SIZE)
                        if not chunk:
                            break
                        f.write(chunk)
                    input_pipe.close()

                self._input_pipe = FileWrapper(
                    io.BufferedReader(io.FileIO(self._tmp_file, "r"))
                )

        self._process = (
            command
            if isinstance(command, subprocess.Popen)
            else self.create_subprocess(command)
        )
        # we want to keep a circular reference to avoid garbage collection
        # when the object is used in, e.g., pipe.read()
        self._process._selfref = self

    def create_subprocess(self, command):
        """
        http://www.chiark.greenend.org.uk/ucgi/~cjwatson/blosxom/2009-07-02-python-sigpipe.html
        """

        def subprocess_setup():
            # Python installs a SIGPIPE handler by default. This is usually not what
            # non-Python subprocesses expect.
            signal.signal(signal.SIGPIPE, signal.SIG_DFL)

        return subprocess.Popen(
            command,
            stdin=self._input_pipe,
            stdout=subprocess.PIPE,
            preexec_fn=subprocess_setup,
            close_fds=True,
        )

    def _finish(self):
        # Need to close this before input_pipe to get all SIGPIPE messages correctly
        self._process.stdout.close()
        if not self._original_input and os.path.exists(self._tmp_file):
            os.remove(self._tmp_file)

        if self._input_pipe is not None:
            self._input_pipe.close()

        self._process.wait()  # deadlock?
        if self._process.returncode not in (0, 141, 128 - 141):
            # 141 == 128 + 13 == 128 + SIGPIPE - normally processes exit with 128 + {reiceived SIG}
            # 128 - 141 == -13 == -SIGPIPE, sometimes python receives -13 for some subprocesses
            raise RuntimeError(
                "Error reading from pipe. Subcommand exited with non-zero exit status %s."
                % self._process.returncode
            )

    def close(self):
        self._finish()

    def __del__(self):
        self._finish()

    def __enter__(self):
        return self

    def _abort(self):
        """
        Call _finish, but eat the exception (if any).
        """
        try:
            self._finish()
        except KeyboardInterrupt:
            raise
        except BaseException:
            pass

    def __exit__(self, type, value, traceback):
        if type:
            self._abort()
        else:
            self._finish()

    def __getattr__(self, name):
        if name in ["_process", "_input_pipe"]:
            raise AttributeError(name)
        try:
            return getattr(self._process.stdout, name)
        except AttributeError:
            return getattr(self._input_pipe, name)

    def __iter__(self):
        for line in self._process.stdout:
            yield line
        self._finish()

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return False


class OutputPipeProcessWrapper(object):
    WRITES_BEFORE_FLUSH = 10000

    def __init__(self, command, output_pipe=None):
        self.closed = False
        self._command = command
        self._output_pipe = output_pipe
        self._process = subprocess.Popen(
            command, stdin=subprocess.PIPE, stdout=output_pipe, close_fds=True
        )
        self._flushcount = 0

    def write(self, *args, **kwargs):
        self._process.stdin.write(*args, **kwargs)
        self._flushcount += 1
        if self._flushcount == self.WRITES_BEFORE_FLUSH:
            self._process.stdin.flush()
            self._flushcount = 0

    def writeLine(self, line):
        assert "\n" not in line
        self.write(line + "\n")

    def _finish(self):
        """
        Closes and waits for subprocess to exit.
        """
        if self._process.returncode is None:
            self._process.stdin.flush()
            self._process.stdin.close()
            self._process.wait()
            self.closed = True

    def __del__(self):
        if not self.closed:
            self.abort()

    def __exit__(self, type, value, traceback):
        if type is None:
            self.close()
        else:
            self.abort()

    def __enter__(self):
        return self

    def close(self):
        self._finish()
        if self._process.returncode == 0:
            if self._output_pipe is not None:
                self._output_pipe.close()
        else:
            raise RuntimeError("Error when executing command %s" % self._command)

    def abort(self):
        self._finish()

    def __getattr__(self, name):
        if name in ["_process", "_output_pipe"]:
            raise AttributeError(name)
        try:
            return getattr(self._process.stdin, name)
        except AttributeError:
            return getattr(self._output_pipe, name)

    def readable(self):
        return False

    def writable(self):
        return True

    def seekable(self):
        return False


class BaseWrapper(object):
    def __init__(self, stream, *args, **kwargs):
        self._stream = stream
        try:
            super(BaseWrapper, self).__init__(stream, *args, **kwargs)
        except TypeError:
            pass

    def __getattr__(self, name):
        if name == "_stream":
            raise AttributeError(name)
        return getattr(self._stream, name)

    def __enter__(self):
        self._stream.__enter__()
        return self

    def __exit__(self, *args):
        self._stream.__exit__(*args)

    def __iter__(self):
        try:
            for line in self._stream:
                yield line
        finally:
            self.close()


class IOPipeline(object):
    """
    Interface for format specifications.
    """

    @classmethod
    def pipe_reader(cls, input_pipe):
        raise NotImplementedError()

    @classmethod
    def pipe_writer(cls, output_pipe):
        raise NotImplementedError()

    def __rshift__(a, b):
        return ChainFormat(a, b)


class ChainFormat(IOPipeline):
    def __init__(self, *args, **kwargs):
        self.args = args
        try:
            self.input = args[0].input
        except AttributeError:
            pass
        try:
            self.output = args[-1].output
        except AttributeError:
            pass
        if not kwargs.get("check_consistency", True):
            return
        for x in range(len(args) - 1):
            try:
                if args[x].output != args[x + 1].input:
                    raise TypeError(
                        "The format chaining is not valid, %s expect %s"
                        " but %s provide %s"
                        % (
                            args[x + 1].__class__.__name__,
                            args[x + 1].input,
                            args[x].__class__.__name__,
                            args[x].output,
                        )
                    )
            except AttributeError:
                pass

    def pipe_reader(self, input_pipe):
        for x in reversed(self.args):
            input_pipe = x.pipe_reader(input_pipe)
        return input_pipe

    def pipe_writer(self, output_pipe):
        for x in reversed(self.args):
            output_pipe = x.pipe_writer(output_pipe)
        return output_pipe


class NopPipeline(IOPipeline):
    def pipe_reader(self, input_pipe):
        return input_pipe

    def pipe_writer(self, output_pipe):
        return output_pipe


class WrappedFormat(IOPipeline):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def pipe_reader(self, input_pipe):
        return self.wrapper_cls(input_pipe, *self.args, **self.kwargs)

    def pipe_writer(self, output_pipe):
        return self.wrapper_cls(output_pipe, *self.args, **self.kwargs)



from dbnd._core.utils.task_utils import to_targets
from dbnd._core.utils.traversing import flatten
from targets.multi_target import MultiTarget


def data_combine(inputs, sort=False):
    targets = flatten(to_targets(inputs))
    if sort:
        targets = sorted(targets, key=lambda x: x.path)
    data = MultiTarget(targets)
    return data



from __future__ import absolute_import

import io

from matplotlib import figure

from targets.marshalling.marshaller import Marshaller
from targets.target_config import FileFormat


class MatplotlibFigureMarshaller(Marshaller):
    type = figure.Figure
    file_format = FileFormat.jpeg

    def value_to_target(self, value, target, **kwargs):
        buf = io.BytesIO()
        value.savefig(buf, **kwargs)

        with target.open("wb") as fd:
            fd.write(buf.getvalue())



# -*- coding: utf-8 -*-
#
# Copyright 2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
The abstract :py:class:`Target` class.
It is a central concept of databand and represents the state of the workflow.
"""

import abc
import logging
import typing

from typing import Any, Dict, Optional

import six

from dbnd._core.utils.basics.nothing import NOTHING


if typing.TYPE_CHECKING:
    from dbnd._core.task.task import Task
    from targets.value_meta import ValueMeta

logger = logging.getLogger(__name__)


class TargetSource(object):
    def __init__(self, task_id, parameter_name=None, name=None):
        self.task_id = task_id
        self.parameter_name = parameter_name
        self.name = name


@six.add_metaclass(abc.ABCMeta)
class Target(object):
    """
    A Target is a resource generated by a :py:class:`~dbnd.tasks.Task`.

    For example, a Target might correspond to a file in HDFS or data in a database. The Target
    interface defines one method that must be overridden: :py:meth:`exists`, which signifies if the
    Target has been created or not.

    Typically, a :py:class:`~dbnd.tasks.Task` will define one or more Targets as output, and the Task
    is considered complete if and only if each of its output Targets exist.
    """

    def __init__(self, properties=None, source=None):
        # type: (Dict[str, Any], TargetSource)->None
        super(Target, self).__init__()
        self.properties = properties or {}
        self.source = source

        # caching and value preview
        self.target_meta = NOTHING  # type: ValueMeta
        self._cache = {}

    @abc.abstractmethod
    def exists(self):
        """
        Returns ``True`` if the :py:class:`Target` exists and ``False`` otherwise.
        """
        pass

    def exist_after_write_consistent(self):
        return True

    @property
    def name(self):
        if self.source_parameter:
            return self.source_parameter.name
        return None

    @property
    def source_task(self):
        # type: ()->Optional[Task]
        if self.source:
            from dbnd._core.current import get_task_by_task_id

            return get_task_by_task_id(self.source.task_id)
        return None

    @property
    def task(self):
        # type: ()->Optional[Task]
        return self.source_task

    @property
    def source_parameter(self):
        if self.source:
            return self.source_task._params.get_param(self.source.parameter_name)
        return None

    def clear_cache(self):
        self._cache = {}

    def set_upstream(self, t):
        self.task.set_upstream(t)

    def set_downstream(self, t):
        self.task.set_downstream(t)

    def __lshift__(self, other):
        return self.set_upstream(other)

    def __rshift__(self, other):
        return self.set_downstream(other)



# Copyright 2017 Stefan Richthofer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created on 13.12.2016
import typing

from typing import Union


def get_Union_params(un):
    """Python version independent function to obtain the parameters
    of a typing.Union object.
    Tested with CPython 2.7, 3.5, 3.6 and Jython 2.7.1.
    """
    try:
        return un.__union_params__
    except AttributeError:
        # Python 3.6
        return un.__args__


def is_Union(tp):
    """Python version independent check if a type is typing.Union.
    Tested with CPython 2.7, 3.5, 3.6 and Jython 2.7.1.
    """
    if tp is Union:
        return True
    try:
        # Python 3.6
        return tp.__origin__ is Union
    except AttributeError:
        try:
            return isinstance(tp, typing.UnionMeta)
        except AttributeError:
            return False


def is_Tuple(tp):
    try:
        return isinstance(tp, typing.TupleMeta)
    except AttributeError:
        try:
            return isinstance(tp, typing._GenericAlias) and tp.__origin__ is tuple
        except AttributeError:
            return False


def get_Tuple_params(tpl):
    """Python version independent function to obtain the parameters
    of a typing.Tuple object.
    Omits the ellipsis argument if present. Use is_Tuple_ellipsis for that.
    Tested with CPython 2.7, 3.5, 3.6 and Jython 2.7.1.
    """
    try:
        return tpl.__tuple_params__
    except AttributeError:
        try:
            if tpl.__args__ is None:
                return None
            # Python 3.6
            if tpl.__args__[0] == ():
                return ()
            else:
                if tpl.__args__[-1] is Ellipsis:
                    return tpl.__args__[:-1] if len(tpl.__args__) > 1 else None
                else:
                    return tpl.__args__
        except AttributeError:
            return None



import logging
import os
import re
import sys

from dbnd._core.current import is_verbose
from dbnd._core.utils.basics.text_banner import safe_string
from dbnd._core.utils.project.project_fs import databand_lib_path, project_path


logger = logging.getLogger(__name__)

_inner_call_functions_regexp = re.compile(
    "|".join(
        [
            r"errors/errors_utils",
            r"_core/task_build",
            r"_core/decorator",
            r"_core/failures",
        ]
    )
)


def safe_value(value):
    if value is None:
        return value

    return safe_string(str(value), 500)


def get_help_msg(cur_ex):
    if getattr(cur_ex, "help_msg", None):
        return cur_ex.help_msg
    return ""


def get_user_frame_info_str(cur_ex):
    if getattr(cur_ex, "user_frame_info", None):
        return frame_info_to_str(cur_ex.user_frame_info)
    return ""


def show_exc_info(cur_ex):
    if not cur_ex:
        return 0
    if hasattr(cur_ex, "show_exc_info"):
        return getattr(cur_ex, "show_exc_info")

    # By default, show exception
    return 1


def _nested_exceptions_messages(ex, limit=None, current=""):
    nested_exceptions = getattr(ex, "nested_exceptions", [])

    if limit:
        limit -= 1
    messages = []

    from dbnd._core.utils.basics.helpers import indent

    for idx, cur_ex in enumerate(nested_exceptions, start=1):
        if cur_ex is None:
            continue
        ex_id = "{current}{idx}".format(current=current, idx=idx)
        message = "{ex_id}. {ex_type}: {msg} ".format(
            ex_id=ex_id, ex_type=type(cur_ex).__name__, msg=str(cur_ex)
        )
        help_msg = get_help_msg(cur_ex)
        if help_msg:
            message += "\n{help_msg}".format(
                help_msg=indent(help_msg, " " * len(ex_id))
            )
        user_frame_info_str = get_user_frame_info_str(cur_ex)
        if user_frame_info_str:
            message += "\n{user_frame_info}".format(
                user_frame_info=indent(user_frame_info_str, " " * len(ex_id))
            )

        messages.append(message)
        if limit is None or limit:
            messages.extend(
                _nested_exceptions_messages(
                    cur_ex, limit=limit, current="(%s)-> " % ex_id
                )
            )
    return messages


def nested_exceptions_str(ex, limit=None):
    messages = _nested_exceptions_messages(ex, limit=limit)

    return "\n".join(messages)


def frame_info_to_str(frame_info):
    if not frame_info:
        return ""
    result = []
    if frame_info.function == "build_dbnd_task":
        return ""
    result.append(
        '  File "{}", line {}, in {}\n'.format(
            frame_info.filename, frame_info.lineno, frame_info.function
        )
    )
    if frame_info.code_context:
        line = "\n".join(l.rstrip() for l in frame_info.code_context)
        result.append("    {}\n".format(line))
    return "".join(result)


def user_side_code(context=5):
    try:
        fi = UserCodeDetector.build_code_detector().find_user_side_frame(
            context=context
        )
        return frame_info_to_str(fi)
    except Exception:
        return "UNKNOWN"


class UserCodeDetector(object):
    def __init__(self, code_dir=None, system_code_dirs=None):
        # The directory we're interested in.
        if not code_dir:
            code_dir = os.getcwd()

        self._system_code_dirs = system_code_dirs or []

        self._system_code_dir = [s + os.path.sep for s in self._system_code_dirs]
        self._code_dir = code_dir

    def is_user_file(self, file):
        """
        Decide whether the file in the traceback is one in our code_dir or not.
        """

        if "site-packages" in file or "dist-packages" in file:
            return False

        if "dbnd_examples" in file:
            return True

        for system_code_dir in self._system_code_dirs:
            if file.startswith(system_code_dir):
                return False
        if file.startswith(self._code_dir):
            return True

        if sys.platform != "win32" and not file.startswith("/"):
            return True

    def _is_user_frame(self, frame_info, user_side_only=True):
        if not frame_info.filename:
            return False

        if _inner_call_functions_regexp.search(frame_info.filename.replace("\\", "/")):
            return False

        if not user_side_only:
            return True

        return self.is_user_file(frame_info.filename)

    def find_user_side_frame(self, depth=1, context=1, user_side_only=False):
        from inspect import getframeinfo

        frame = sys._getframe(depth)
        while frame:
            frame_info = getframeinfo(frame, context)
            if self._is_user_frame(frame_info, user_side_only):
                return frame_info

            frame = frame.f_back
        return None

    @classmethod
    def build_code_detector(cls, system_code_dirs=None):
        # should be called withing databand environment

        if system_code_dirs is None:
            system_code_dirs = [databand_lib_path(), databand_lib_path("../targets")]
        code_dir = project_path()
        return cls(code_dir=code_dir, system_code_dirs=system_code_dirs)


def log_exception(msg, ex, logger_=None, verbose=None, non_critical=False):
    logger_ = logger_ or logger

    from dbnd._core.errors.base import DatabandError

    if verbose is None:
        verbose = is_verbose()

    if verbose:
        # just show the exception
        logger_.exception(msg)
        return

    if non_critical:
        logger_.info(msg + ": %s" % str(ex))
        return

    if isinstance(ex, DatabandError):
        # msg = "{msg}:{ex}".format(msg=msg, ex=ex)
        logger_.error(msg + ": %s" % str(ex))
    else:
        # should we? let's show the exception for now so we can debug
        logger_.exception(msg)



from __future__ import absolute_import

import sys

from dbnd._core.errors import (
    DatabandBuildError,
    DatabandConfigError,
    DatabandError,
    DatabandRuntimeError,
    TaskClassAmbigiousException,
)
from dbnd._core.errors.errors_utils import safe_value
from dbnd._core.errors.friendly_error import (
    api,
    config,
    execute_engine,
    executor_k8s,
    graph,
    targets,
    task_build,
    task_execution,
    task_parameters,
    task_registry,
    tools,
)
from dbnd._core.errors.friendly_error.helpers import (
    _band_call_str,
    _run_name,
    _safe_target,
    _safe_task_family,
)
from dbnd._core.utils.basics.format_exception import format_exception_as_str


def task_found_in_airflow_dag(root_task):
    return DatabandError(
        "Task '%s' implementation has been discovered via DAGs loading" % root_task,
        help_msg="Your task %s were loaded via dags folder, currently we don't support that.\n"
        "You should define your tasks not in airflow/dag folder, but inside your project.\n"
        "Use module key in a [databand] section of config file ( $DBND__DATABAND__MODULE )"
        % root_task,
        show_exc_info=False,
    )


def task_has_not_complete_but_all_outputs_exists(task):
    return DatabandRuntimeError(
        "Something wrong, task %s has been executed, "
        "but _complete function returns False while all outputs exist! "
        % _safe_task_family(task),
        help_msg="Check your implementation of %s, validate that you _complete override is correct",
        show_exc_info=False,
    )


def task_has_missing_outputs_after_execution(task, missing_str):
    return DatabandRuntimeError(
        "Task %s has been executed, but some outputs are missing!\n\t%s"
        % (_safe_task_family(task), missing_str),
        help_msg="Check your implementation of %s, validate that all outputs has been written. "
        "If it's directory, validate that you have _SUCCESS flag "
        "(self.your_task_output.mark_success())" % _run_name(task),
        show_exc_info=False,
    )


def task_data_source_not_exists(task, missing, downstream=None):
    if downstream:
        tasks = ",".join(map(str, downstream))
        dependent = "Tasks that depend on this input are: %s\n" % tasks
    else:
        dependent = ""
    if len(missing) == 1:
        missing_target = missing[0]

        from targets.dir_target import DirTarget

        if (
            isinstance(missing_target, DirTarget)
            and missing_target.folder_exists()
            and missing_target.flag_target
        ):
            # we are missing flag!
            return DatabandRuntimeError(
                "Data source '%s' success flag is missing! %s"
                % (missing_target.flag_target, dependent),
                help_msg="Check that SUCCESS flag exists and your configurations is ok. "
                "You can override flag check by \n"
                "1. adding '[noflag]' to your path: '%s[noflag]' \n"
                "2. --PARAMETER-target '[noflag]' \n"
                "3. Define parameter using 'parameter.folder.with_flag(None)' \n"
                "4 .Create the flag if you think that the input is ok" % missing_target,
                show_exc_info=False,
            )
        return DatabandRuntimeError(
            "Task input at location '%s' is missing! %s" % (missing_target, dependent),
            help_msg="Validate that this data exists. "
            "This string considered as a path, as it defined as 'data' in our system",
            show_exc_info=False,
        )

    if len(missing) > 5:
        missing_msg = "%s... (%s files)" % (missing[:10], len(missing))
    else:
        missing_msg = ",".join(map(str, missing))

    return DatabandRuntimeError(
        "Data source '%s' is missing! %s" % (missing_msg, dependent),
        help_msg="Check that file exists and your configurations is ok. "
        "If it's directory, validate that you have _SUCCESS flag, "
        "or override that via target config ('noflag')",
        show_exc_info=False,
    )


def ambiguous_task(full_task_name):
    return TaskClassAmbigiousException(
        "Task %r is ambiguous, you have more than one task with the same name!"
        % full_task_name,
        help_msg="Check whether this module contains more than one methods with the same name. "
        "Please use full reference (like 'a.b.c') ",
        show_exc_info=False,
    )


def no_matching_tasks_in_pipeline(tasks, tasks_regexes):
    all_tasks_names = ",".join([t.task_id for t in tasks])
    return DatabandConfigError(
        "None of '%s' tasks have been found at current pipeline!" % tasks_regexes,
        help_msg="check your run.selected_tasks_regex switch, "
        "select one of following tasks: %s" % all_tasks_names,
        show_exc_info=False,
    )


def dag_with_different_contexts(task_id):
    return DatabandRuntimeError(
        "The task '%s' isn't part of the current context!" % task_id,
        help_msg="The task '%s' isn't part of the current context! \n"
        "Make sure you did not fiddle with internal APIs" % task_id,
        show_exc_info=False,
    )


def dbnd_module_not_found_tip(module):
    if str(module).startswith("dbnd"):
        dbnd_module = module.split(".")[0]
        return "Do you have '%s' installed?" % (dbnd_module.replace("_", "-"))
    return ""


def failed_to_import_user_module(ex, module, description):
    s = format_exception_as_str(sys.exc_info())
    msg = "Module '%s' can not be loaded: %s" % (
        module,
        dbnd_module_not_found_tip(module),
    )
    return DatabandError(
        "%s exception: %s." % (msg, s),
        help_msg=" Databand is trying to load user module '%s' as required by %s: \n "
        "Probably, it has compile errors or not exists." % (module, description),
        show_exc_info=False,
    )


def unknown_args_in_task_call(parent_task, cls, func_params):
    return DatabandBuildError(
        "You are trying to create %s from %s with *args, please use named arguments only: args=%s"
        % (_run_name(cls), _band_call_str(parent_task), func_params[0]),
        show_exc_info=True,
        help_msg="Check your %s logic" % _band_call_str(parent_task),
    )


def failed_to_create_task(exc_desc, nested_exceptions):
    msg = "Failed to create task %s" % exc_desc
    if nested_exceptions:
        msg += ": "
        msg += ",".join([str(e) for e in nested_exceptions[:3]])
        if len(nested_exceptions) > 3:
            msg += " (showing 3 errors out of %s)" % len(nested_exceptions)
    return DatabandBuildError(
        msg, show_exc_info=False, nested_exceptions=nested_exceptions
    )


def failed_to_convert_value_to_target(value):
    return DatabandBuildError(
        "Can't convert '%s' of type %s to target." % (safe_value(value), type(value)),
        help_msg="You can convert only values with type of "
        "Target, Tasks and strings to the Path/PathStr/Target",
    )


def failed_to_calculate_task_parameter_value(ex, task_family, p_name, exc_desc):
    return DatabandBuildError(
        "Failed to process parameter %s for %s: %s" % (p_name, exc_desc, ex),
        nested_exceptions=[ex],
        help_msg="Please validate your config/commandline/.band that creates runs %s"
        % exc_desc,
    )


def failed_to_read_pandas(ex, target):
    return DatabandError(
        "There is an error while reading {target}: {ex}".format(
            target=_safe_target(target), ex=ex
        ),
        nested_exceptions=[ex],
    )


def failed_to_set_index(ex, df, set_index, target):
    return DatabandError(
        "Failed to set index to '{set_index}' "
        "for data frame with columns {columns} "
        "while reading from {target}: {ex}".format(
            set_index=set_index, columns=df.columns, target=_safe_target(target), ex=ex
        ),
        nested_exceptions=[ex],
    )


def failed_to_read_target_as_task_input(ex, task, parameter, target):
    return DatabandError(
        "Failed to read '{task.task_name}.{p.name}' from "
        "'{target}' as format '{target.config}' to {p.value_type.type_str}: {ex}"
        "".format(p=parameter, target=target, task=task, ex=ex),
        nested_exceptions=ex,
    )


def failed_to_read_task_input(ex, task, parameter, target):
    return DatabandError(
        "Failed to read '{task.task_name}.{p.name}' from "
        "'{target}' to {p.value_type.type_str}: {ex}"
        "".format(p=parameter, target=target, task=task, ex=ex),
        nested_exceptions=ex,
    )


def failed_to_write_task_output(ex, target, value_type):
    return DatabandError(
        "Failed to write '{target.name}' to '{target} {target.config}  ({value_type}): {ex}"
        "".format(value_type=value_type, target=target, ex=ex),
        nested_exceptions=ex,
    )


def failed_to_write_pandas(ex, target):
    return DatabandError(
        "There is an error while writing to {target} {ex}".format(
            target=_safe_target(target), ex=ex
        ),
        nested_exceptions=ex,
    )


def marshaller_no_merge(marshaller, target, partitions):
    return DatabandError(
        "Can't merge {p_len} partitions on read from {target} "
        "as current marshaller {marshaller} doesn't support merge funcitonality".format(
            p_len=len(partitions), target=target, marshaller=marshaller
        )
    )


def airflow_bad_user_configuration(ex, file_path):
    return DatabandConfigError(
        "Error while trying to load additional airflow configuration from %s"
        % file_path,
        help_msg="Please make sure that the configuration file %s does exist."
        % file_path,
        nested_exceptions=ex,
        show_exc_info=False,
    )


def airflow_versioned_dag_missing(command):
    return DatabandError(
        "Could not run '%s', dbnd-airflow-versioned-dag is not installed." % command,
        help_msg="Please run 'pip install dbnd-airflow-versioned-dag' in order to run '%s'."
        % command,
        show_exc_info=False,
    )



import datetime
import enum

import attr
import six

from dbnd._core.utils.timezone import make_aware, utcnow


RESULT_PARAM = "result"

CURRENT_DATETIME = utcnow()
CURRENT_TIME_STR = CURRENT_DATETIME.strftime("%Y%m%d_%H%M%S")
CURRENT_DATE = CURRENT_DATETIME.date()


class EnvLabel(object):  # env label
    dev = "dev"
    test = "test"
    staging = "stage"
    qa = "qa"
    prod = "prod"


class CloudType(object):
    local = "local"
    gcp = "gcp"
    aws = "aws"
    azure = "azure"


class TaskExecutorType(object):
    local = "local"


class OutputMode(object):
    regular = "regular"
    prod_immutable = "prod_immutable"


class _ConfigParamContainer(object):
    _type_config = True

    @classmethod
    def is_type_config(self, cls):
        # we can't use issubclass as there are "generic" types that will fail this check
        return getattr(cls, "_type_config", False)


class _TaskParamContainer(object):
    pass


class DescribeFormat(object):
    short = "short"
    long = "long"
    verbose = "verbose"


# Compute Types


class EnumWithAll(enum.Enum):
    @classmethod
    def all(cls):
        return list(cls)

    @classmethod
    def all_values(cls):
        return [x.value for x in cls]

    @classmethod
    def all_names(cls):
        return [x.name for x in cls]


class SparkClusters(EnumWithAll):
    local = "local"
    dataproc = "dataproc"
    databricks = "databricks"
    emr = "emr"
    qubole = "qubole"


class ApacheBeamClusterType(object):
    local = "local"
    dataflow = "dataflow"


class ClusterPolicy(object):
    NONE = "none"
    CREATE = "create"
    KILL = "kill"
    EPHERMAL = "ephermal"

    ALL = [NONE, CREATE, KILL, EPHERMAL]


class EmrClient(object):
    LIVY = "livy"
    STEP = "step"


class TaskType(object):
    pipeline = "pipeline"
    python = "python"
    spark = "spark"
    pyspark = "pyspark"
    dataflow = "dataflow"
    docker = "docker"


class TaskRunState(EnumWithAll):
    SCHEDULED = "scheduled"
    QUEUED = "queued"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"

    CANCELLED = "cancelled"
    SHUTDOWN = "shutdown"

    UPSTREAM_FAILED = "upstream_failed"
    SKIPPED = "skipped"
    UP_FOR_RETRY = "up_for_retry"

    @staticmethod
    def final_states():
        return TaskRunState.finished_states() | {
            TaskRunState.UPSTREAM_FAILED,
            TaskRunState.SKIPPED,
        }

    @staticmethod
    def final_states_str():
        return [s.value for s in TaskRunState.final_states()]

    @staticmethod
    def finished_states():
        return {TaskRunState.SUCCESS, TaskRunState.FAILED, TaskRunState.CANCELLED}

    @staticmethod
    def direct_fail_states():
        return {TaskRunState.FAILED, TaskRunState.CANCELLED}

    @staticmethod
    def fail_states():
        return {
            TaskRunState.FAILED,
            TaskRunState.CANCELLED,
            TaskRunState.UPSTREAM_FAILED,
        }

    @staticmethod
    def states_lower_case():
        return [state.name.lower() for state in TaskRunState]


REUSED = "reused"


class RunState(EnumWithAll):
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"
    SHUTDOWN = "shutdown"
    CANCELLED = "cancelled"


class AlertStatus(EnumWithAll):
    TRIGGERED = "TRIGGERED"
    RESOLVED = "RESOLVED"
    ACKNOWLEDGED = "ACKNOWLEDGED"


class AlertErrorPolicy(object):
    none = ""
    all = "all"
    task_only = "task_only"


class SystemTaskName(object):
    driver_submit = "dbnd_driver_submit"
    driver = "dbnd_driver"
    task_submit = "dbnd_task_submit"

    driver_and_submitter = {driver_submit, driver}


@attr.s
class _DbndDataClass(object):
    def asdict(self, filter=None):
        return attr.asdict(self, recurse=False, filter=filter)


HEARTBEAT_DISABLED = make_aware(datetime.datetime.fromtimestamp(0))


class ParamValidation(EnumWithAll):
    warn = "warn"
    error = "error"
    disabled = "disabled"


class DbndTargetOperationType(EnumWithAll):
    init = "init"
    read = "read"
    write = "write"
    reuse = "reuse"
    log = "log"
    log_hist = "log_hist"


class DbndTargetOperationStatus(EnumWithAll):
    OK = "OK"
    NOK = "NOK"


class SystemMetrics(EnumWithAll):
    Duration = "Duration"
    TotalCpuTime = "Total CPU Time"
    TotalWallTime = "Total Wall Time"
    ColdTotalCpuTime = "Cold Total CPU Time"
    ColdTotalWallTime = "Cold Total Wall Time"

    @staticmethod
    def duration_metrics():
        """Used to select metrics for removal during metrics re-generation"""
        return [
            s.value
            for s in [
                SystemMetrics.Duration,
                SystemMetrics.TotalCpuTime,
                SystemMetrics.TotalWallTime,
                SystemMetrics.ColdTotalCpuTime,
                SystemMetrics.ColdTotalWallTime,
            ]
        ]


class UpdateSource(EnumWithAll):
    dbnd = "dbnd"
    airflow_monitor = "airflow_monitor"
    airflow_tracking = "airflow_tracking"
    azkaban_tracking = "azkaban_tracking"

    def __eq__(self, other):
        if isinstance(other, UpdateSource):
            return self.value == other.value
        elif isinstance(other, six.string_types):
            return str(self) == other or str(self.value) == other

        return False

    @classmethod
    def is_tracking(cls, source):
        return source in [UpdateSource.airflow_tracking, UpdateSource.azkaban_tracking]


class MetricSource(object):
    user = "user"
    system = "system"
    histograms = "histograms"
    spark = "spark"

    @classmethod
    def all(cls):
        return [cls.user, cls.system, cls.histograms, cls.spark]

    @classmethod
    def default_sources(cls):
        return [cls.user, cls.system, cls.histograms, cls.spark]

    @classmethod
    def default_sources_str(cls):
        return ",".join(cls.default_sources())


class AlertSeverity(object):
    CRITICAL = "CRITICAL"
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"

    @classmethod
    def values(cls):
        return [cls.CRITICAL, cls.HIGH, cls.MEDIUM, cls.LOW]



import typing

from typing import Any

import attr


if typing.TYPE_CHECKING:
    from dbnd._core.parameter.parameter_definition import ParameterDefinition


@attr.s(slots=True)
class ParameterValue(object):
    parameter = attr.ib()  # type: ParameterDefinition
    value = attr.ib()  # type: Any

    source = attr.ib()  # type: str
    source_value = attr.ib()  # type: Any

    parsed = attr.ib(default=True)  # type: bool
    warnings = attr.ib(factory=list)

    @property
    def name(self):
        return self.parameter.name



import os

from targets.base_target import Target
from targets.caching import DbndLocalFileMetadataRegistry
from targets.data_target import DataTarget
from targets.dir_target import DirTarget
from targets.errors import (
    FileAlreadyExists,
    FileSystemException,
    MissingParentDirectory,
    NotADirectory,
)
from targets.file_target import FileSystemTarget, FileTarget
from targets.fs.file_system import FileSystem
from targets.fs.local import LocalFileSystem
from targets.inmemory_target import InMemoryTarget
from targets.target_factory import target
from targets.utils.atomic import AtomicLocalFile



import re

from dbnd._vendor.termcolor import colored


# todo: those are const - should be calculate at run time (so colored function could could be removed if needed)
ERROR_SEPARATOR = colored("--------------------------------------", color="red")
ERROR_SEPARATOR_SMALL = colored("- - -", on_color="on_red")


def bold(value):
    return colored(value, attrs=["bold"])


COLOR_REGEX = re.compile(r"\x1b?\[(\d;)?\d*m")


def uncolor(text):
    """removes the ANSII color formatting from text"""
    return COLOR_REGEX.sub("", text)



import logging
import os.path
import tarfile
import tempfile

from dbnd import output, parameter, task
from dbnd._core.constants import CloudType
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.plugin.dbnd_plugins import assert_web_enabled
from dbnd._core.task import config
from dbnd._core.utils.timezone import utcnow
from targets import Target
from targets.types import Path


logger = logging.getLogger(__name__)


class DbSyncConfig(config.Config):
    """(Advanced) Databand's db sync behaviour"""

    _conf__task_family = "db_sync"

    export_root = parameter[Target]


@task(archive=output(output_ext=".tar.gz")[Path])
def export_db(
    archive,
    include_db=True,
    include_logs=True,
    task_version=utcnow().strftime("%Y%m%d_%H%M%S"),
):
    # type: (Path, bool, bool, str)-> None

    from dbnd._core.current import get_databand_context

    logger.info("Compressing files to %s..." % archive)
    with tarfile.open(str(archive), "w:gz") as tar:

        if include_db:

            assert_web_enabled(reason="dbnd_web is required for export db")
            dbnd_context = get_databand_context()
            conn_string = dbnd_context.settings.web.get_sql_alchemy_conn()
            if conn_string.startswith("sqlite:"):
                from dbnd_web.utils.dbnd_db import get_sqlite_db_location

                db_file = get_sqlite_db_location(conn_string)
                logger.info("Exporting DB=%s", db_file)
                tar.add(db_file, arcname="dbnd.db")
            elif conn_string.startswith("postgresql"):
                with tempfile.NamedTemporaryFile(prefix="dbdump.", suffix=".sql") as tf:
                    from dbnd_web.utils.dbnd_db import dump_postgres

                    dump_postgres(conn_string, tf.name)
                    tar.add(tf.name, arcname="postgres-dbnd.sql")
            else:
                raise DatabandRuntimeError(
                    "Can not export db! "
                    "Currently, we support only sqlite and postgres db in automatic export"
                )

        if include_logs:
            context = get_databand_context()
            local_env = context.settings.get_env_config(CloudType.local)
            logs_folder = local_env.dbnd_local_root.folder("logs").path
            if os.path.exists(logs_folder):
                logger.info("Adding run folder from '%s'", logs_folder)
                tar.add(logs_folder, "run")
            else:
                logger.error("Logs dir '%s' is not found", logs_folder)



import json
import os
import re

from datetime import datetime, timedelta
from typing import Any, Dict

import attr
import six

from dbnd._core.current import try_get_databand_run
from targets.config import is_in_memory_cache_target_value


@attr.s(frozen=True)
class TargetCacheKey(object):
    target = attr.ib(converter=str)
    value_type = attr.ib(converter=str)


class TargetCache(object):
    def __init__(self, cache=None):
        self._cache = cache if cache is not None else {}

    def get_cache_group(self):
        # type: () -> Dict[TargetCacheKey, Any]
        return self._cache

    def set(self, value, key):
        if not self.enabled:
            return

        cache = self.get_cache_group()
        if cache is not None:
            cache[key] = value

    def get(self, key):
        if not self.enabled:
            return

        cache = self.get_cache_group()
        if cache is not None:
            return cache.get(key)

    def has(self, key):
        cache = self.get_cache_group()
        if cache is not None:
            return key in cache
        return False

    def __getitem__(self, item):
        return self.get(key=item)

    def __setitem__(self, key, value):
        return self.set(value=value, key=key)

    def __contains__(self, item):
        return self.has(key=item)

    def clear_for_targets(self, targets_to_clear):
        if not targets_to_clear:
            return

        cache = self.get_cache_group()
        if not cache:
            return

        targets_to_clear = set(targets_to_clear)
        for key in list(cache.keys()):
            if key.target in targets_to_clear and key in cache:
                del cache[key]

    def clear(self):
        cache = self.get_cache_group()
        if cache:
            cache.clear()

    def clear_all(self):
        self._cache.clear()

    @property
    def enabled(self):
        return is_in_memory_cache_target_value()


TARGET_CACHE = TargetCache()


class DbndLocalFileMetadataRegistry(object):
    """
    :type file_path: str
    :type local_md5: str
    :type remote_md5: str
    :type created_at: datetime.datetime
    :type ttl: str
    """

    ext = ".dbnd-meta"
    _date_format = "%Y-%m-%d %H:%M:%S"
    default_ttl = "1d"

    def __init__(
        self, file_path, local_md5=None, remote_md5=None, created_at=None, ttl=None,
    ):
        self._file_path = file_path
        self._cache_file_path = self._resolve_cache_file_name(file_path)
        self._local_md5 = local_md5
        self._remote_md5 = remote_md5
        self._created_at = self._resolve_created_at(created_at)

        self._ttl = ttl or DbndLocalFileMetadataRegistry.default_ttl
        if not self.is_valid_ttl(self._ttl):
            raise Exception(
                "Valid ttl should be inf form of 1-9*d|1-9*h, is %s".format(self._ttl)
            )

    def __eq__(self, other):
        return all(
            [
                self._file_path == other._file_path,
                self._cache_file_path == other._cache_file_path,
                self._local_md5 == other._local_md5,
                self._remote_md5 == other._remote_md5,
                self._created_at == other._created_at,
            ]
        )

    def __repr__(self):
        return "DbndFileCache<{}:{}:{}>".format(
            self._file_path, self._created_at, self._ttl
        )

    @staticmethod
    def _resolve_created_at(created_at):
        if not created_at:
            return datetime.now()
        if six.PY2:
            created_at = str(created_at)
        if isinstance(created_at, str):
            return datetime.strptime(
                created_at, DbndLocalFileMetadataRegistry._date_format
            )
        if isinstance(created_at, datetime):
            return created_at

        raise Exception(
            "`created_at` parameter must be either datetime object or isoformat date string"
        )

    @staticmethod
    def is_valid_ttl(ttl):
        """Check if provided ttl value is in supported format i.e 1h, 12h, 2d"""
        return bool(re.match(r"\d*d$|\d*h$", ttl))

    @staticmethod
    def _resolve_cache_file_name(file_path):
        run = try_get_databand_run()
        if not run:
            raise Exception("No databand run found to when creating cache file")
        dbnd_local_root = run.get_current_dbnd_local_root()
        cache_dir = get_or_create_folder_in_dir("cache", dbnd_local_root.path)

        file_name = os.path.basename(file_path) + DbndLocalFileMetadataRegistry.ext
        return os.path.join(cache_dir, file_name)

    @classmethod
    def read(cls, file_target):
        """Read file cache and create new instance of DbndFileCache"""
        cache_file_path = cls._resolve_cache_file_name(file_target.path)
        with open(cache_file_path, "r+") as f:
            kwargs = json.loads(f.read())
        return cls(**kwargs)

    def save(self):
        """Save DbndFileCache instance to json-like cache file"""
        data = {
            "file_path": self._file_path,
            "local_md5": self._local_md5,
            "remote_md5": self._remote_md5,
            "created_at": self._created_at.strftime(
                DbndLocalFileMetadataRegistry._date_format
            ),
            "ttl": self._ttl,
        }
        with open(self._cache_file_path, "w+") as f:
            f.write(json.dumps(data))

    @staticmethod
    def exists(file_target):
        """Read existing cache file, if it not exists then return None"""
        cache_file_path = DbndLocalFileMetadataRegistry._resolve_cache_file_name(
            file_target.path
        )
        return os.path.exists(cache_file_path)

    def delete(self):
        """Remove cache file"""
        if os.path.isfile(self._cache_file_path):
            os.remove(self._cache_file_path)

    @property
    def expired(self):
        """Check if cache file is still valid"""
        time_map = {
            "d": timedelta(days=1),
            "h": timedelta(hours=1),
        }
        interval_type = self._ttl[-1]
        interval_value = int(self._ttl[:-1])
        expired_at = self._created_at + interval_value * time_map[interval_type]
        return expired_at < datetime.now()

    def validate_local_md5(self, md5_hash):
        return self._local_md5 == md5_hash

    @staticmethod
    def get_or_create(file_target):
        dbnd_meta_cache_exists = DbndLocalFileMetadataRegistry.exists(file_target)
        if dbnd_meta_cache_exists:
            return DbndLocalFileMetadataRegistry.read(file_target)
        else:
            dbnd_meta_cache = DbndLocalFileMetadataRegistry(file_path=file_target.path)
            dbnd_meta_cache.save()
            return dbnd_meta_cache

    @staticmethod
    def refresh(file_target):
        dbnd_meta_cache = DbndLocalFileMetadataRegistry.get_or_create(file_target)
        if dbnd_meta_cache.expired:
            dbnd_meta_cache.delete()
            return DbndLocalFileMetadataRegistry.get_or_create(file_target)


def get_or_create_folder_in_dir(folder_name, dir_):
    # type: (str, Task) -> str
    folder_path = os.path.join(dir_, folder_name)
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)
    return folder_path



from dbnd._core.task.config import Config
from dbnd._core.task.data_combine_task import data_combine
from dbnd._core.task.data_source_task import DataSourceTask, data_source
from dbnd._core.task.pipeline_task import PipelineTask
from dbnd._core.task.python_task import PythonTask
from dbnd._core.task.task import Task
from dbnd.tasks import basics


__all__ = [
    "Task",
    "Config",
    "data_combine",
    "data_source",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
]

try:
    from dbnd_spark.spark import PySparkTask, SparkTask, spark_task

    __all__ += ["SparkTask", "PySparkTask", "spark_task"]
except ImportError:
    pass
